"""
NOTE: These methods operate directly on the kernel matrix!!!
some functions are adapted from Felix Musil's ml_tools
"""

import numpy as np

from .base import RegressorBase
from .getscore import get_score

class KRR(RegressorBase):

    def __init__(self, jitter):
        # Weights of the krr model
        self.alpha = None
        self.jitter = jitter  # noise level^2
        self.coninv = None  # inverse of the covariance matrix
        self._fitted = False

    def fit(self, kernel, y, kNM=[]):
        '''Train the krr model with trainKernel and trainLabel.'''

        reg = np.eye(kernel.shape[0]) * self.jitter
        self.coninv = np.linalg.inv(kernel + reg)
        self.alpha = np.linalg.solve(kernel + reg, y)
        self._fitted = True

    def predict(self, kernel):
        '''kernel.shape is expected as (nPred, nTrain)'''
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use predict.")
        return np.dot(kernel, self.alpha.flatten()).reshape((-1))

    def predict_uncertainty(self, k, delta):
        '''
        k.shape is expected as (nPred, nTrain), delta is the variance of y
        '''
        n_k = len(k)
        y_error = np.zeros(n_k)
        for i in range(n_k):
            y_error[i] = np.sqrt(delta * (1. - np.dot(k[i], np.dot(self.coninv, k[i]))))
        return y_error

    def get_params(self, deep=True):
        return dict(sigma=self.jitter)

    def set_params(self, params, deep=True):
        self.jitter = params['jitter']
        self.alpha = None

    def pack(self):
        state = dict(weights=self.alpha, jitter=self.jitter)
        return state

    def unpack(self, state):
        self.alpha = state['weights']
        err_m = 'jitter are not consistent {} != {}'.format(self.jitter, state['jitter'])
        assert self.jitter == state['jitter'], err_m

    def loads(self, state):
        self.alpha = state['weights']
        self.jitter = state['jitter']


class KRRSparse(RegressorBase):

    def __init__(self, jitter, delta, sigma):
        # Weights of the krr model
        self.alpha = None
        self.jitter = jitter
        self.delta = None  # variance of the prior
        self.sigma = sigma  # noise
        self._fitted = False
        if self.jitter is None:
            self.jitter = 10e-20

    def fit(self, kMM, y, kNM):
        '''N train structures, M sparsified representative structures '''
        '''kMM: the kernel matrix of the representative structures with shape (M,M)'''
        '''kNM: the kernel matrix between the representative and the train structures with shape (N,M)'''

        # if (kMM.shape[0] != kMM.shape[1]):# or kMM.shape[0] != kNM.shape[1] or kNM.shape[0] != y.shape[0]):
        # raise ValueError('Shape of the kernel matrix is not consistent!')

        if self.delta is None:
            self.delta = np.std(y) / (np.trace(kMM) / len(kMM))

        if self.sigma is None:
            self.sigma = 0.001 * np.std(y)

        sparseK = kMM * self.delta * self.sigma ** 2 + np.dot(kNM.T, kNM) * self.delta ** 2
        sparseY = np.dot(kNM.T, y)
        reg = np.eye(kMM.shape[0]) * self.jitter

        self.alpha = np.linalg.solve(sparseK + reg, sparseY)
        self._fitted = True

    def predict(self, kNM):
        '''kNM: the kernel matrix between the representative and the new structures with shape (N,M)'''
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use predict.")
        return np.dot(self.delta ** 2 * kNM, self.alpha.flatten()).reshape((-1))

    def fit_predict(self, kMM, kNM, y, kNM_test):
        self.fit(kMM, kNM, y)
        return self.predict(kNM_test)

    def fit_predict_error(self, kMM, kNM, y, kNM_test, y_test):
        self.fit(kMM, kNM, y)
        return self.predict_error(kNM_test, y_test)

    def get_params(self, deep=True):
        return dict(jitter=self.jitter, delta=self.delta, sigma=self.sigma)

    def set_params(self, params, deep=True):
        self.jitter = params['jitter']
        self.sigma = params['sigma']
        self.alpha = None

    def pack(self):
        state = dict(weights=self.alpha, jitter=self.jitter, delta=self.delta, sigma=self.sigma)
        return state

    def unpack(self, state):
        self.alpha = state['weights']
        self.delta = state['delta']
        self.sigma = state['sigma']
        err_m = 'jitter are not consistent {} != {}'.format(self.jitter, state['jitter'])
        assert self.jitter == state['jitter'], err_m

    def loads(self, state):
        self.alpha = state['weights']
        self.jitter = state['jitter']
        self.delta = state['delta']
        self.sigma = state['sigma']


class KRRFastCV(RegressorBase):
    """ 
    taken from:
    An, S., Liu, W., & Venkatesh, S. (2007). 
    Fast cross-validation algorithms for least squares support vector machine and kernel ridge regression. 
    Pattern Recognition, 40(8), 2154-2162. https://doi.org/10.1016/j.patcog.2006.12.015
    """

    def __init__(self, jitter, delta, cv):
        self.jitter = jitter
        self.cv = cv
        self.delta = delta
        self._fitted = False

    def fit(self, kernel, y, kNM=[]):
        '''Fast cv scheme. Destroy kernel.'''
        np.multiply(self.delta ** 2, kernel, out=kernel)
        kernel[np.diag_indices_from(kernel)] += self.jitter
        kernel = np.linalg.inv(kernel)
        alpha = np.dot(kernel, y)
        Cii = []
        beta = np.zeros(alpha.shape)
        self.y_pred = np.zeros(y.shape)
        self.error = np.zeros(y.shape)
        for _, test in self.cv.split(kernel):
            Cii = kernel[np.ix_(test, test)]
            beta = np.linalg.solve(Cii, alpha[test])
            self.y_pred[test] = y[test] - beta
            self.error[test] = beta  # beta = y_true - y_pred

        del kernel
        self._fitted = True

    def predict(self, kernel=None):
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use predict.")
        '''kernel.shape is expected as (nPred, nTrain)'''
        return self.y_pred

    def get_params(self, deep=True):
        return dict(sigma=self.jitter, cv=self.cv)

    def set_params(self, params, deep=True):
        self.jitter = params['jitter']
        self.cv = params['cv']
        self.delta = params['delta']
        self.y_pred = None

    def pack(self):
        state = dict(y_pred=self.y_pred, cv=self.cv.pack(),
                     jitter=self.jitter, delta=self.delta)
        return state

    def unpack(self, state):
        self.y_pred = state['y_pred']
        self.cv.unpack(state['cv'])
        self.delta = state['delta']

        err_m = 'jitter are not consistent {} != {}'.format(self.jitter, state['jitter'])
        assert self.jitter == state['jitter'], err_m

    def loads(self, state):
        self.y_pred = state['y_pred']
        self.cv.loads(state['cv'])
        self.jitter = state['jitter']
        self.delta = state['delta']



"""
some operations for kernel and distance matrices
"""

import copy

import numpy as np


def normalizekernel(kernel):
    # first normalize the kernel matrix
    nkernel = copy.deepcopy(kernel)
    size = len(kernel)
    for i in range(size):
        nkernel[i, :] /= np.sqrt(kernel[i, i])
        nkernel[:, i] /= np.sqrt(kernel[i, i])
        nkernel[i, i] = 1.0
    return nkernel.clip(max=1)


def kerneltodis(kernel):
    # there can be many transformations between the k-matrix and the distance matrix
    # Here we use d_ij = sqrt(2 - 2*k_ij)
    # (k_ij is a normalized symmetric kernel)
    nk = normalizekernel(kernel)
    size = len(kernel)
    dis = np.zeros((size, size), dtype=np.float64)
    for i in range(size):
        for j in range(i - 1):
            dis[i, j] = dis[j, i] = np.sqrt(2. - 2. * nk[i, j])

    return dis.clip(min=0)


def kerneltodis_linear(kernel):
    # there can be many transformations between the k-matrix and the distance matrix
    # Here we use d_ij = 1-k_ij
    # (k_ij is a normalized symetric kernel)
    nk = normalizekernel(kernel)
    dis = 1. - nk
    return dis.clip(min=0)


def kerneltorho(kernel, delta):
    # we compute the "density" of the data from kernel matrix
    # delta is the charecteristic spread in similarity

    rho = np.zeros(len(kernel))
    allrhofromdis = np.exp((np.asmatrix(kernel) - 1.0) / delta)

    for i in range(len(allrhofromdis)):
        rho[i] = np.sum(allrhofromdis[i])

    return rho


def distorho_quick(dis, delta):
    # we compute the "density" of the data from distance matrix
    # the distance matrix can be computed such as
    # dis = kerneltodis(kernel)
    # delta is the charecteristic distance
    rho = np.zeros(len(dis))
    allrhofromdis = np.exp(dis * (-1. / delta))

    for i in range(len(allrhofromdis)):
        rho[i] += np.sum(allrhofromdis[i])

    return rho



#!/usr/bin/python
import sys
import argparse
import numpy as np
from asaplib.hypers.univeral_length_scales import uni_length_scales, system_pair_bond_lengths

"""
Automatically generate the hyperparameters of ACSF descriptors for arbitrary elements and combinations.

## Example
The command
gen_universal_soap_hypers.py --Zs 1

generates ACSF for hydrogen
"""

def main(Zs, scalerange, sharpness):

    """
    scalerange: type=float, scale the cutoffs of the SFs.
    sharpness: type=float, sharpness factor for atom_width, default=1.0, 
                larger sharpness means more resolution, and more SFs will be generated.
    rmin: type=float, distance in Angstrom to the first nearest neighbor. 
         Eliminates the symmetry functions that investigate the space between 0 and rmin.
    """
    for Z in Zs:
        if str(Z) not in uni_length_scales:
            raise RuntimeError("key Z {} not present in length_scales table".format(Z))

    shortest_bond, longest_bond = system_pair_bond_lengths(Zs, uni_length_scales)
    cutoff = longest_bond * 1.56 * float(scalerange)
    rmin = shortest_bond
    N = min(int(sharpness*(cutoff-rmin)/0.5), 5)

    index = np.arange(N+1, dtype=float)
    shift_array = cutoff*(1./N)**(index/(len(index)-1)) 
    eta_array = 1./shift_array**2.

    for fel in Zs:
        for sel in Zs:
            print("# symfunctions for type %s 2 %s" %(fel, sel))
            for eta in eta_array:
                if 3*np.sqrt(1/eta) > rmin:
                    print("symfunction_short %s 2 %s %.4f 0.000 %.3f" %(fel, sel, eta, cutoff))
            for i in range(len(shift_array)-1):
                eta = 1./((shift_array[N-i] - shift_array[N-i-1])**2)
                if shift_array[N-i] + 3*np.sqrt(1/eta) > rmin:
                    print("symfunction_short %s 2 %s %.4f %.3f %.3f" %(fel, sel, eta, shift_array[N-i], cutoff))

    eta_array = np.logspace(-3,0,N//2)
    zeta_array = [1.000, 4.000, 16.000]
    for fel in Zs:
        ang_Zs = list(Zs)
        for sel in Zs:
            for tel in ang_Zs:
                print("# symfunctions for type %s 3 %s %s" %(fel, sel, tel))
                for eta in eta_array:
                    for zeta in zeta_array:
                        if 3*np.sqrt(1/eta) > rmin:
                            print("symfunction_short %s 3 %s %s %.4f  1.000 %.3f %.3f" %(fel, sel, tel, eta, zeta, cutoff))
                            print("symfunction_short %s 3 %s %s %.4f -1.000 %.3f %.3f" %(fel, sel, tel, eta, zeta, cutoff))
            ang_Zs.pop(0)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=None)
    parser.add_argument("--Zs", nargs="+", type=int, help="atomic numbers to calculate descriptors for", required=True)
    parser.add_argument('-sr', '--scalerange', type = float, default = 1.0, help = 'Scale the length scales.')
    parser.add_argument('-s', '--sharpness', type = float, default = 1.0, help = 'larger means more ACSFs will be selected.')
    args = parser.parse_args()
    main(args.Zs, args.scalerange, args.sharpness)



import click

from asaplib.io import ConvertStrToList


def state_input_options(f):
    """Create common options for I/O files"""
    f = click.option('--in_file', '--in', '-i', type=click.Path('r'),
                     help='The state file that includes a dictionary-like specifications of descriptors to use.')(f)
    return f


def file_input_options(f):
    """Create common options for I/O files"""
    f = click.option('--fxyz', '-f',
                     type=str,
                     help='Input file that contains XYZ coordinates.\
                           See a list of possible input formats:\
                          https://wiki.fysik.dtu.dk/ase/ase/io/io.html\
                          If a wildcard * is used, all files matching the pattern is read.',
                     default=None)(f)
    return f


def file_input_format_options(f):
    """Create common options for I/O files"""
    f = click.option('--fxyz_format',
                     type=str,
                     help='Additional info for the input file format. e.g.\
                          {"format":"lammps-data","units":"metal","style":"full"}',
                     default=None)(f)
    return f


def file_output_options(f):
    """Create common options for I/O files"""
    f = click.option('--prefix', '-p',
                     help='Prefix to be used for the output file.',
                     default=None)(f)
    return f


def dm_input_options(f):
    """common options for reading a design matrices, used for map, fit, kde, clustering, etc."""
    f = click.option('--design_matrix', '-dm', cls=ConvertStrToList, default='[]',
                     help='Location of descriptor matrix file or name of the tags in ase xyz file\
                           the type is a list  \'[dm1, dm2]\', as we can put together simutanously several design matrix.')(
        f)
    f = click.option('--use_atomic_descriptors', '--use_atomic', '-ua',
                     help='Use atomic descriptors instead of global ones.',
                     default=False, is_flag=True)(f)
    f = click.option('--only_use_species', type=int,
                     help='Only use the atomic descriptors of species with the specified atomic number.\
                           Only makes sense if already using --use_atomic_descriptors.',
                     default=None)(f)
    return f


def km_input_options(f):
    """common options for reading a kernel matrices, can be used for map, fit, kde, clustering, etc."""
    f = click.option('--kernel_matrix', '-km', default='none',
                     help='Location of a kernel matrix file')(f)
    return f


def output_setup_options(f):
    """Create common options for output results from clustering/KDE analysis"""
    f = click.option('--savexyz/--no-savexyz',
                     help='Save the results to the xyz file',
                     default=True)(f)
    f = click.option('--savetxt/--no-savetxt',
                     help='Save the results to the txt file',
                     default=False)(f)
    return f


def desc_options(f):
    """Create common options for computing descriptors"""
    f = click.option('--tag',
                     help='Tag for the descriptors.',
                     default='cmd-desc')(f)
    return f


def para_options(f):
    """Create common options for parallellization"""
    f = click.option('--number_processes', '--nprocess', '-np', type=int,
                     help='Number of processes when compute the descriptors in parrallel.',
                     show_default=True, default=1)(f)
    return f


def atomic_to_global_desc_options(f):
    """Create common options for global descriptors constructed based on atomic fingerprints """
    f = click.option('--reducer_type', '-r',
                     help='type of operations to get global descriptors from the atomic soap vectors, e.g. \
                          [average], [sum], [moment_avg], [moment_sum].',
                     show_default=True, default='average', type=str)(f)
    f = click.option('--zeta', '-z',
                     help='Moments to take when converting atomic descriptors to global ones.',
                     default=1, type=int)(f)
    f = click.option('--element_wise', '-e',
                     help='element-wise operation to get global descriptors from the atomic soap vectors',
                     show_default=True, default=False, is_flag=True)(f)
    f = click.option('--peratom', '-pa',
                     help='Save the per-atom local descriptors.',
                     show_default=True, default=False, is_flag=True)(f)
    return f


def map_setup_options(f):
    """Create common options for making 2D maps of the data set"""
    f = click.option('--peratom',
                     help='Save the per-atom projection.',
                     default=False, is_flag=True)(f)
    f = click.option('--adjusttext/--no-adjusttext',
                     help='Adjust the annotation texts so they do not overlap.',
                     default=False)(f)
    f = click.option('--annotate', '-a',
                     help='Location of tags to annotate the samples.',
                     default='none', type=str)(f)
    f = click.option('--aspect_ratio', '-ar',
                     help='Aspect ratio of the plot',
                     show_default=True, default=2, type=float)(f)
    f = click.option('--style', '-s',
                     type=click.Choice(['default', 'journal'], case_sensitive=False),
                     help='Style of the plot.',
                     show_default=True, default='default')(f)
    return f


def map_io_options(f):
    """Create common options for making 2D maps of the data set"""
    f = click.option('--keepraw/--no-keepraw',
                     help='Keep the high dimensional descriptor when output XYZ file.',
                     default=False)(f)
    f = click.option('--output', '-o', type=click.Choice(['xyz', 'matrix', 'none', 'chemiscope'], case_sensitive=False),
                     help='Output file format.',
                     default='xyz')(f)
    f = click.option('--extra-properties', '-ep', type=click.Path(exists=True),
                     help='Additional properties to be read for each frmae in CSV format.')(f)
    return f


def color_setup_options(f):
    """Create common options for handing color scales"""
    f = click.option('--normalized_by_size', '-nbs',
                     help='Normalize the quantity used for color function by the number of atoms in each frame.',
                     show_default=True, default=False, is_flag=True)(f)
    f = click.option('--colormap', '-cmap',
                     help='Colormap used. Common options: gnuplot, tab10, viridis, bwr, rainbow.',
                     show_default=True, default='gnuplot')(f)
    f = click.option('--color_from_zero', '-c0',
                     help='Set the minimum to zero and only plot the excess.',
                     show_default=True, default=False, is_flag=True)(f)
    f = click.option('--color_label', '-clab',
                     help='The label for the color bar.',
                     default=None)(f)
    f = click.option('--color_column', '-ccol',
                     help='The column number used in the color file. Starts from 0.',
                     default=0)(f)
    f = click.option('--color', '-c',
                     help='Location of a file or name of the properties in the XYZ file. \
                     Used to color the scatter plot for all samples (N floats).',
                     default='none', type=str)(f)
    return f


def d_reduce_options(f):
    """Create common options for dimensionality reduction"""
    f = click.option('--axes', nargs=2, type=click.Tuple([int, int]),
                     help='Plot the projection along which projection axes.',
                     default=[0, 1])(f)
    f = click.option('--dimension', '-d',
                     help='Number of the dimensions to keep in the output XYZ file.',
                     default=10)(f)
    f = click.option('--scale/--no-scale',
                     help='Standard scaling of the coordinates.',
                     default=True)(f)
    return f


def fit_setup_options(f):
    """Create common options for making 2D maps of the data set"""
    f = click.option('--lc_points', '-lcp', type=int,
                     help='the number of sub-samples to take when compute the learning curve',
                     show_default=True, default=8)(f)
    f = click.option('--learning_curve', '-lc', type=int,
                     help='the number of points on the learning curve, <= 1 means no learning curve',
                     show_default=True, default=-1)(f)
    f = click.option('--test_ratio', '--test', '-t', type=float,
                     help='Test ratio.',
                     show_default=True, default=0.05)(f)
    f = click.option('--normalized_by_size', '-nbs',
                     help='Normalize y by the number of atoms in each frame.',
                     show_default=True, default=False, is_flag=True)(f)
    f = click.option('--y', '-y',
                     help='Location of a file or name of the properties in the XYZ file',
                     default='none', type=str)(f)
    return f


def kernel_options(f):
    """Create common options for compute kernel functions"""
    f = click.option('--kernel_parameter', '-kp', type=float,
                     help='Parameter used in the kernel function.',
                     default=None)(f)
    f = click.option('--kernel', '-k',
                     type=click.Choice(['linear', 'polynomial', 'cosine'], case_sensitive=False),
                     help='Kernel function for converting design matrix to kernel matrix.',
                     show_default=True, default='linear')(f)
    return f


def sparsification_options(f):
    """Create common options for sparsification"""
    f = click.option('--n_sparse', '-n', type=int,
                     help='number of the representative samples, set negative if using no sparsification',
                     show_default=True, default=100)(f)
    f = click.option('--sparse_mode', '-s',
                     type=click.Choice(['random', 'cur', 'fps', 'sequential'], case_sensitive=False),
                     help='Sparsification method to use.',
                     show_default=True, default='fps')(f)
    return f



"""
class and methods for performing kernel density estimation
"""
import numpy as np
from scipy.stats import gaussian_kde
from sklearn.neighbors import KernelDensity

class Kernel_Density_Base:
    """
    Base class for performing kernel density estimation
    """
    def __init__(self):
        self.acronym = 'none'

    def fit(self, X):

        """Fit kernel model to X"""

        pass

    def evaluate_density(self, X):
        """Given an array of data, computes the local density of every point using kernel density estimation

        Input
        ------
        Data X : array, shape(n_sample,n_feature)

        Return
        ------
        Log of densities for every point: array, shape(n_sample)
        """
        pass

    def fit_evaluate_density(self, X):
        self.fit(X)
        return self.evaluate_density(X)

    def get_acronym(self):
        # we use an acronym for each KDE, so it's easy to find it and refer to it
        return self.acronym

class KDE_scipy(Kernel_Density_Base):
    """
    Kernel Density Estimation with Scipy
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html
    """
    def __init__(self, bw_method=None):
        """
        bw_method: str, scalar or callable, optional

        The method used to calculate the estimator bandwidth. 
        This can be ‘scott’, ‘silverman’, a scalar constant or a callable. 
        If a scalar, this will be used directly as kde.factor. 
        If a callable, it should take a gaussian_kde instance as only parameter and return a scalar. 
        If None (default), ‘scott’ is used. 
        """
        self.bw_method = bw_method
        self.kde = None
        self.acronym = 'kde_scipy'
        self._fitted = False

    def fit(self, X):
        """
        X: dataset, array_like

        Datapoints to estimate from. In case of univariate data this is a 1-D array, 
        otherwise a 2-D array with shape (# of data, # of dimension)

        Note that scipy.stats.gaussian_kde take X with shape (# of dimension, # of data)
        This is why we transpose the input X.
        """
        if isinstance(self.bw_method, float):
            # Note that scipy weights its bandwidth by the covariance of the
            # input data.  To make the results comparable to the other methods,
            # we divide the bandwidth by the sample standard deviation here.
            self.kde = gaussian_kde(X.T, bw_method=self.bw_method/x.std(ddof=1))
        else:
            self.kde = gaussian_kde(X.T, bw_method=self.bw_method)
        self._fitted = True

    def evaluate_density(self, X):
        if self._fitted is False:
            raise ValueError("The KDE model has not been fitted.")
        return np.log(self.kde.evaluate(X.T))
        

class KDE_sklearn(Kernel_Density_Base):
    """
    Kernel Density Estimation with Sklearn
    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KernelDensity.html#sklearn.neighbors.KernelDensity
    https://scikit-learn.org/stable/modules/density.html#kernel-density-estimation
    """
    def __init__(self, bandwidth=1.0, algorithm='auto', kernel='gaussian', metric='euclidean'):
        """
        bandwidth: float. The bandwidth of the kernel.
        algorithm: str. The tree algorithm to use. Valid options are [‘kd_tree’|’ball_tree’|’auto’]. Default is ‘auto’.
        kernel:str. The kernel to use. 
                    Valid kernels are [‘gaussian’|’tophat’|’epanechnikov’|’exponential’|’linear’|’cosine’] 
                    Default is ‘gaussian’.
        metric: str. The distance metric to use. 
        """
        self.bandwidth = bandwidth
        self.algorithm = algorithm
        self.kernel = kernel
        self.metric = metric
        self.kde = KernelDensity(bandwidth=bandwidth, algorithm=algorithm, kernel=kernel, metric=metric)
        self.acronym = 'kde_sklearn'
        self._fitted = False

    def fit(self, X):
        """
        X: dataset, array_like

        Datapoints to estimate from. In case of univariate data this is a 1-D array,
        otherwise a 2-D array with shape (# of data, # of dimension)

        """
        self.kde.fit(X)
        self._fitted = True

    def evaluate_density(self, X):
        if self._fitted is False:
            raise ValueError("The KDE model has not been fitted.")
        return self.kde.score_samples(X)



"""
Density-based Clustering Algorithms
"""

from matplotlib import pyplot as plt
import numpy as np
import scipy as sp
import json
from yaml import dump as ydump
from yaml import Dumper
from sklearn.cluster import DBSCAN
from sklearn.metrics import silhouette_score
from sklearn.neighbors import NearestNeighbors

from .ml_cluster_base import ClusterBase, FitClusterBase
from ..io import NpEncoder

class DBCluster(ClusterBase):
    """
    Performing clustering using density based clustering algorithm
    """
    _pairwise = True

    def __init__(self, trainer):
        super().__init__()
        self.trainer = trainer
        # cluster labels
        self.labels = None
        # number of clusters
        self.n_clusters = None
        # number of noise points
        self.n_noise = None

    def fit(self, dmatrix, rho=None):
        """fit the clustering model, assume input of NxN distance matrix or Nxm coordinates"""

        self.labels = self.trainer.fit(dmatrix, rho)
        # Number of clusters in labels, ignoring noise if present.
        self.n_clusters = len(set(self.labels)) - (1 if -1 in self.labels else 0)
        self.n_noise = list(self.labels).count(-1)
        print('Estimated number of clusters: %d' % self.n_clusters)
        print('Estimated number of noise points: %d' % self.n_noise)

        if np.shape(dmatrix)[0] == np.shape(dmatrix)[1]:
            silscore = silhouette_score(dmatrix, self.labels, metric="precomputed")
        else:
            silscore = silhouette_score(dmatrix, self.labels, metric="euclidean")
        print("Silhouette Coefficient: %0.3f" % silscore)

    def get_cluster_labels(self, index=[]):
        """return the label of the samples in the list of index"""
        if len(index) == 0:
            return self.labels
        else:
            return self.labels[index]

    def get_n_cluster(self):
        return self.n_clusters

    def get_n_noise(self):
        return self.n_noise

    def pack(self):
        """return all the info"""
        state = dict(trainer=self.trainer.name, trainer_params=self.trainer.pack(), labels=self.labels.tolist(),
                     n_clusters=self.n_clusters, n_noise=self.n_noise)
        return state

    def save_state(self, filename, mode='json'):
         if mode == 'yaml':
             with open(filename+'-clustering-state.yaml', 'w') as yd:
                 ydump(self.pack(), yd, sort_keys=True, Dumper=Dumper)
         else:
             with open(filename+'-clustering-state.json', 'w') as jd:
                 json.dump(self.pack(), jd, sort_keys=True, cls=NpEncoder)

class sklearn_DB(FitClusterBase):

    """
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
    eps : float, optional
        The maximum distance between two samples for one to be considered as in the neighborhood of the other.
        This is not a maximum bound on the distances of points within a cluster.
        This is the most important DBSCAN parameter to choose appropriately for your dataset and distance function.

    min_samples : int, optional
        The number of samples (or total weight) in a neighborhood for a point to be considered as a core point.
        This includes the point itself.

    metric : string, or callable
        The metric to use when calculating distance between instances in a feature array.
        If metric is a string or callable, it must be one of the options allowed by sklearn.metrics.pairwise_distances for its metric parameter.
        If metric is “precomputed”, X is assumed to be a distance matrix and must be square.
        X may be a sparse matrix, in which case only “nonzero” elements may be considered neighbors for DBSCAN.
    """
    _pairwise = True

    def __init__(self, eps=None, min_samples=None, metrictype='precomputed'):
        super().__init__()
        self.name='DBSCAN'
        self.metric = metrictype  # e.g. 'euclidean'
        # distance metric
        self.eps = eps
        # The number of samples in a neighborhood for a point to be considered as a core point.
        self.min_samples = min_samples
        self.db = None

    def fit(self, dmatrix, rho=None):
        self.db = DBSCAN(eps=self.eps, min_samples=self.min_samples, metric=self.metric).fit(dmatrix)
        return self.db.labels_

    def pack(self):
        """return all the info"""
        return self.db.get_params()


class LAIO_DB(FitClusterBase):
    """
    Clustering by fast search and find of density peaks, Rodriguez and Laio (2014).

    https://science.sciencemag.org/content/sci/344/6191/1492.full.pdf

    :math: \rho_i\,=\,\sum_j{\chi(d_{ij}-d_{cut})}
    :math: \delta_i\,=\,\min_{j:\rho_j>\rho_i}(d_{ij})
    """
    def __init__(self, distances=None, indices=None, dens_type=None, dc=None, percent=2.0):

        """
        Parameters
        ----------
        distances: numpy array 
        array of distances between data points and their neighbours of shape (Nele, n_neigh_comp)
        where Nele is the number of data points and n_neigh_comp is the number of neighbours to consider.

        indices: numpy array 
               array of shape (Nele, n_neigh_comp) where row i gives the indices of the neighbours for data point i

        dens_type: str
                 The type of density to compute. Can be 'exp' (exponential) or None (linear)

        dc: float 
        giving the cutoff distance beyond which data points don't contribute to the local density computation
        of another datapoint.

        percent: int
        Criterion for choosing dc, int from 0-100, typically chosen such that the average number of neighbours
        is 1-2% of the total number of points in the dataset. Default is 2%.
        """

        self.name='LAIO_DB'
        self.distances = distances
        self.indices = indices
        self.dens_type = dens_type
        self.dc = dc
        self.percent = percent

        # numpy array of shape (Nele,) where the densities of the data points expressed in terms of the number of data
        # points within the cutoff distance
        self.dens = None

        # numpy array of shape (Nele,) of the distances to the nearest data point to i
        # possessing a higher local density.
        self.delta_to_neigh = None

        # numpy array of shape (Nele,) where the ith entry gives the index of the nearest neighbour with
        # higher density than data point i.
        self.ref_neigh = None

        # Unused currently
        self.decision_graph = None

        # Clusters must lie at a distance greater than self.delta_cut (float) apart to be designated as independent
        # clusters
        self.delta_cut = None

        # If a "cluster" has a density smaller than self.dens_cut (float) it is discarded as noise
        self.dens_cut = None

        # numpy array of shape (Nele, ) giving the cluster (int from 0 to N) each data point belongs to.
        # Clusters are labelled by density. Cluster 0 has the highest density and cluster N has the smallest density.
        self.cluster = None

        # numpy array of shape (N,) giving the indices of the N cluster centres.
        self.center_indices = None

        # numpy array of shape (Nele,) where halo points are designated as -1 and otherwise are assigned to their
        # respective cluster centres
        self.halo = None

    def get_dc(self, data):
        """
        Compute the cutoff distance given the data.

        Parameters
        ----------
        data: np.matrix 
        The data array of shape (Nele, proj_dim) where N is the number of data points and proj_dim is the number of
        projected components of the kernel matrix.

        Returns
        -------
        self.dc: float
             the cutoff distance
        """

        Nele = data.shape[0]

        # The number of neighbours to consider is dictated by the percent parameter provided (default is 2% of the
        # total data points.
        n_neigh = int(self.percent/100.*Nele)

        # Compute with 4 times the recommended percentage of neighbours in order that we may re-estimate the percent
        # value supplied (rper) based on the data.
        n_neigh_comp = int(4.*self.percent/100.*Nele)
        neigh = NearestNeighbors(n_neighbors=n_neigh_comp).fit(data)
        self.distances, self.indices = neigh.kneighbors(data)

        # The cutoff distance dc is the average distance (averaged over all data points) of the "0.02*Nele"th neighbour
        dc = np.average(self.distances[:, n_neigh])

        # To store the local densities of each data point
        dens = np.zeros(data.shape[0])
        tt = 1.0
        factor = 1.0

        # We update dc based on a re-estimated percentage rper. Stop when rper is close to self.percent
        while tt > 0.05:
            dc = dc*factor
            for i in range(Nele):
                a = self.distances[i, :]
                dens[i] = len(a[(a <= dc)])
            rper = 100.*np.average(dens)/Nele
            tt = rper/self.percent - 1.
            if tt > 0.:
                factor = factor/1.001
            else:
                factor = factor*1.001
            tt = abs(tt)
        self.dc = dc

    def get_decision_graph(self, data, fplot=True):
        """
        Method currently doesn't produce the decision graph.

        Parameters
        ----------
        data: numpy array of shape (Nele, proj_dim).
        fplot: Boolean indicating whether or not to plot the decision graph

        Returns
        -------

        """

        Nele = data.shape[0]
        self.dens = np.zeros(Nele)

        if self.dc == None:
            self.get_dc(data)

        else:  # check compatibility between the provided cutoff distance and the provided distance matrix
            n_neigh_comp = int(10.*self.percent/100.*Nele)
            neigh = NearestNeighbors(n_neighbors=n_neigh_comp).fit(data)
            self.distances, self.indices = neigh.kneighbors(data)
            if self.dc > np.min(self.distances[:, n_neigh_comp - 1]):
                print("dc too big for being included within the", 10.*self.percent, "%  of data, consider using a "
                                                                                    "small dc or augment the percent "
                                                                                    "parameter")

        for i in range(Nele):
            a = self.distances[i, :]
            self.dens[i] = len(a[(a <= self.dc)])

        if self.dens_type == 'exp':  # Density is no longer number but exponential
            for i in range(Nele):
                a = self.distances[i, :]/self.dc  # distances expressed as a fraction of the cutoff distance
                self.dens[i] = np.sum(np.exp(-a**2))

        # The densities computed by this class or not log densities. Log densities are returned by the KDE class.

        # self.dens_cut = 0.2 * np.mean(np.log(self.dens)) + 0.8 * np.min(np.log(self.dens))
        self.dens_cut = 0.2 * np.mean(self.dens) + 0.8 * np.min(self.dens)
        self.delta_cut = np.mean(self.distances)

        self.delta_to_neigh = np.zeros(data.shape[0])
        self.ref_neigh = np.zeros(data.shape[0], dtype='int')
        indices = np.arange(data.shape[0])
        imax = []  # holds index of data point with highest local density

        for i in range(data.shape[0]):
            higher_dens_indices = indices[((self.dens > self.dens[i]) & (indices != i))]
            higher_dens_data = data[((self.dens > self.dens[i]) & (indices != i))]
            if higher_dens_data.shape[0] > 0:
                ds = np.transpose(sp.spatial.distance.cdist([np.transpose(data[i, :])], higher_dens_data))
                j = np.argmin(ds)
                self.ref_neigh[i] = higher_dens_indices[j]
                self.delta_to_neigh[i] = ds[j]
            else:
                self.delta_to_neigh[i] = -100.
                imax.append(i)
        self.delta_to_neigh[imax] = np.max(self.delta_to_neigh)*1.05

        # Plot kernel density on x-axis and delta_to_neigh (point of higher density) on the y-axis

        if fplot:

            plt.scatter(self.dens, self.delta_to_neigh)
            plt.plot([min(self.dens), max(self.dens)], [self.delta_cut, self.delta_cut], c='red')
            plt.plot([self.dens_cut, self.dens_cut], [np.min(self.distances), np.max(self.distances)], c='red')
            plt.title('Decision Graph')
            plt.xlabel('rho')
            plt.ylabel('delta')
            plt.show()

    def get_assignation(self, data):
        """
        Parameters
        ----------
        data: numpy array of shape (Nele, proj_dim) where proj_dim gives the number of dimensions the kernel matrix is
              projected to
        Returns: self.halo numpy array of shape (Nele, ) where halo points are designated as -1 and otherwise are
        assigned to their respective cluster centres.
        """
        ordered_by_dens = np.argsort(-self.dens)  # data points in descending order of density
        self.cluster = np.zeros(data.shape[0], dtype='int')
        indices = np.arange(data.shape[0])
        center_label = np.zeros(data.shape[0], dtype='int')
        ncluster = -1
        for i in range(data.shape[0]):  # trick to iterate through the ordered_by_dens array
            j = ordered_by_dens[i]
            if (self.dens[j] > self.dens_cut) & (self.delta_to_neigh[j] > self.delta_cut):
                ncluster = ncluster + 1
                self.cluster[j] = ncluster
                center_label[j] = ncluster
            else:
                # the assigned cluster is the cluster the closest neighbour of higher density belongs to
                self.cluster[j] = self.cluster[self.ref_neigh[j]]
                center_label[j] = -1
        self.center_indices = indices[(center_label != -1)]
        # border points is a binary array, 1 if data point has a neighbour that has been assigned a different cluster
        # centre and 0 otherwise.
        border_points = np.zeros(data.shape[0], dtype='int')
        self.halo = np.copy(self.cluster)

        for i in range(data.shape[0]):
            for j in self.indices[i, :][(self.distances[i, :] <= self.dc)]:
                if self.cluster[i] != self.cluster[j]:
                    border_points[i] = 1
        halo_cutoff = np.zeros(ncluster + 1)
        for i in range(ncluster + 1):
            # Extract the densities of a cluster's border points
            dens_of_border_points = self.dens[((border_points == 1) & (self.cluster == i))]
            if len(dens_of_border_points) == 0:  # if a cluster has no border points then it has no halo points.
                halo_cutoff[i] = 0
            else:
                halo_cutoff[i] = np.max(dens_of_border_points)
            i += 1
        self.halo[indices[(self.dens < halo_cutoff[self.cluster])]] = -1

        return self.halo

    def fit(self, data, rho=None):
        """
        Compute the center labels.

        Parameters
        ----------
        data: numpy array of shape (Nele, proj_dim) where proj_dim is the number of components the kernel matrix has been
              projected to.
        rho: densities, default is None since the DP.py module computes the densities itself.
        Returns: cluster_labels: numpy array of shape (Nele,) giving the cluster (int from 0 to N) each data point
        belongs to. Halo points are designated as -1.
        -------
        """

        self.get_dc(data)
        self.get_decision_graph(data)
        cluster_labels = self.get_assignation(data)

        return cluster_labels

    def pack(self):
        """
        Return dictionary containing the cutoff distance, self.dc, for data points to contribute to the local density
        of another data point as well as self.dens_cut, the density threshold for defining a cluster.
        """
        state = dict(deltamin=self.dc, rhomin=self.dens_cut)
        return state


"""
Legacy Code below for old Fast Search and Find of Density Peaks class.
"""


class old_LAIO(FitClusterBase):
    """Laio Clustering scheme
    Clustering by fast search and find of density peaks
    https://science.sciencemag.org/content/sci/344/6191/1492.full.pdf

    :math: \\rho_i\,=\,\\sum_j{\\chi(d_{ij}-d_{cut})}
         i.e. the local density of data point x_i
    :math: \\delta_i\,=\,\\min_{j:\\rho_j>\\rho_i}(d_{ij})
         i.e. the minimum distance to a neighbour with higher density


    A summary of laio clustering algorithm:
    1. First do a kernel density estimation (rho_i) for each data point i
    2. For each data point i, compute the distance (delta_i) between i and j,
       j is the closet data point that has a density higher then i, i.e. rho(j) > rho(i).
    3. Plot the decision graph, which is a scatter plot of (rho_i, delta_i)
    4. Select cluster centers ({cl}), which are the outliers in the decision graph that fulfills:
       i) rho({cl}) > rhomin
       ii) delta({cl}) > delta_min
       one needs to set the two parameters rhomin and delta_min.
    5. After the cluster centers are determined, data points are assigned to the nearest cluster center.
 one needs to set two parameters:
    """
    _pairwise = True

    def __init__(self, deltamin=-1, rhomin=-1):
        """
        :param deltamin: the lower bound on the distance between two cluster centres
        :param rhomin: The lower bound in kernel density, if the density of a cluster is lower
                       than this threshold, this "cluster" will be discarded as noise
        """

        self.deltamin = deltamin
        self.rhomin = rhomin

    def fit(self, dmatrix, rho=None):

        """

        Parameters
        ----------
        dmatrix: The distance matrix of shape (Nele, Nele)
        rho: The log densities of the points of shape (Nele,)

        Returns
        -------

        """

        if rho is None:
            raise ValueError('for fdb it is better to compute kernel density first')

        delta, nneigh = self.estimate_delta(dmatrix, rho)

        #  if there's no input values for rhomin and deltamin, we use simple heuristics
        if self.rhomin < 0:
            self.rhomin = 0.2*np.mean(rho) + 0.8*np.min(rho)
        if self.deltamin < 0:
            self.deltamin = np.mean(delta)

        #here we make the decision graph
        #x axis (rho) is the kernel density for each data point
        #y axis (delta) is the distance to the nearest higher density points

        plt.scatter(rho, delta)
        plt.plot([min(rho), max(rho)], [self.deltamin, self.deltamin], c='red')
        plt.plot([self.rhomin, self.rhomin], [min(delta), max(delta)], c='red')
        plt.xlabel('rho')
        plt.ylabel('delta')
        plt.show()

        nclust = 0  # cluster index
        cl = np.zeros(len(rho), dtype='int')-1  # numpy array of -1's
        for i in range(len(rho)):
            if rho[i] > self.rhomin and delta[i] > self.deltamin:
                nclust += 1
                cl[i] = nclust

        # Assignment
        ordrho = np.argsort(rho)[::-1]  # Indices of data points in descending order of local density
        rho_ord = rho[ordrho]
        for i in range(len(rho)):
            if cl[ordrho[i]] == -1:
                cl[ordrho[i]] = cl[nneigh[ordrho[i]]]
        return cl

    def estimate_delta(self, dist, rho):
        """
        For each data point i, compute the distance (delta_i) between i and j,
        j is the closest data point that has a density higher then i, i.e. rho(j) > rho(i).

        Parameters
        ----------
        dist: distance matrix of shape (Nele, Nele)
        rho: log densities for each data point array of shape (Nele,)

        Returns: delta: numpy array of distances to nearest cluster centre for each datapoint.
                 nneight: numpy array giving the index of the nearest cluster centre.
        -------

        """
        delta = (rho*0.0).copy()
        nneigh = np.ones(len(delta), dtype='int')
        for i in range(len(rho)):
            #  for data i, find all points that have higher density
            js = np.where(rho > rho[i])[0]
            #  if there's no j's that have higher density than i, we set delta_i to be a large distance
            if len(js) == 0:
                delta[i] = np.max(dist[i, :])
                nneigh[i] = i
            else:
                # find the nearest j that has higher density then i
                delta[i] = np.min(dist[i, js])
                nneigh[i] = js[np.argmin(dist[i, js])]
        return delta, nneigh

    def pack(self):
        '''return all the info'''
        state = dict(deltamin=self.deltamin, rhomin=self.rhomin)
        return state



"""
Tools for performing dimensionality reduction of design matrices
using sparse KPCA
"""

import numpy as np
from ..compressor import Sparsifier
from ..kernel import Descriptors_to_Kernels
from .ml_kpca import KernelPCA

class SPARSE_KPCA:
    def __init__(self, n_components=2, kernel={}, sparse_mode="fps", n_sparse=None):
        """
        Parameters
        ----------
        n_components : int, default=2
            number of dimensions to project to

        kernel: dictionaries that specify which way to convert descriptors into kernel matrix
                     See ../kernel/kernel_transforms.py`
        e.g.
        kernel = {
        "first_kernel": {"type": 'linear', "normalize": True}
        }
        e.g.
        {'k0':{"type": "cosine"}} 
        e.g.
        { 'k1': {"type": "polynomial", "d": power}}

        sparsemode: str, default='fps', 
                    possible method to use ([fps], [cur],[random],[sequential])'
        n_sparse: int, number of the representative samples, negative means no sparsification
         
        """

        # essential
        self.n_components = n_components
        self.sparse_mode = sparse_mode
        self.n_sparse = n_sparse
        self.kernel = kernel

        if self.n_sparse is None or self.n_sparse > 0:
            self.sparsifier = Sparsifier(self.sparse_mode)
        else:
            self.sparsifier = None

        # object for transform design matrix to kernel matrix
        self.k_transform = Descriptors_to_Kernels(kernel)
        # object for doing kpca
        self.kpca = KernelPCA(self.n_components)

        # we need to store the representative structures
        self.sbs = None
        # design matrix of the representative structures
        self.desc_sbs = None

        # state properties
        self._fitted = False

    def _sparsify(self, desc):
        # sparsification
        n_sample = len(desc)
        # set default value of n_sparse
        if self.n_sparse is None:
            self.n_sparse = max(10, n_sample // 20)
        # sparsification
        if self.n_sparse > 0:
            self.sbs = self.sparsifier.sparsify(desc, self.n_sparse)
        else:
            print("Not using any sparsification")
            self.sbs = range(n_sample)

        self.desc_sbs = desc[self.sbs]

    def fit(self, desc):
        """Fit KPCA on the precomputed descriptor matrix

        Parameters
        ----------
        desc: array-like, shape=[n_descriptors, n_samples]
        design matrix

        Returns
        -------
        """
        if self._fitted:
            raise RuntimeError('PCA already fitted before, please reinitialise the object!')

        self._sparsify(desc)

        kNN = self.k_transform.compute(self.desc_sbs)
        self.kpca.fit(kNN)

        self._fitted = True

    def transform(self, desc_test):
        """Transforms to the lower dimensions

        Parameters
        ----------
        desc_test : array_like, shape=[n_descriptors, n_samples]
            design matrix of the new samples

        Returns
        -------
        projected_vectors: numpy.array, shape=(L,n_components)
            projections of the design matrix on low dimension
        """
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use transform.")

        kNM = self.k_transform.compute(desc_test, self.desc_sbs)
        return self.kpca.transform(kNM)

    
    def fit_transform(self, desc):
        """Fit PCA on the design matrix & project to lower dimensions

        Parameters
        ----------
        desc : array_like, shape=[n_descriptors, n_samples]
            design matrix of the new samples

        Returns
        -------
        projected_vectors: numpy.array, shape=(L,n_components)
            projections of the design matrix on low dimension

        """
        self.fit(desc)
        return self.transform(desc)



"""
Select samples using a re-weighted distribution

The original distribution (KDE) of the sample is
:math: \rho = exp(-F)
and we select the samples using a well-tempered distribution
:math: \rho(\lambda) = exp(-F/\lambda)  
"""

import numpy as np

def reweight(logkde, n_sparse, reweight_lambda):

    """
    Parameters:
    ------------
    logkde: list, type=float
            The (log of) kernel density for each sample
    reweight_lambda: float
              reweighting factor
    n_sparse: int
              number of samples to select
    
    Return:
    ------------
    sbs: list, type=int
         A list of selected samples
    """
    nframes = len(logkde)

    new_kde = np.zeros(nframes)
    for i in range(nframes):
        new_kde[i] = np.exp(logkde[i] / reweight_lambda) / np.exp(logkde[i])
        # compute the normalization factor so we expect to select n samples in the end
        normalization = n_sparse / np.sum(new_kde)
        new_kde *= normalization
        sbs = []
        randomchoice = np.random.rand(nframes)
        for i in range(nframes):
            if randomchoice[i] < new_kde[i]:
                sbs.append(i)

    return sbs



"""
functions used in cmd_asap.py
"""

"""for loading in_file"""


def load_in_file(in_file):
    """Here goes the routine to compute the descriptors according to the state file(s)"""

    with open(in_file, 'r') as stream:
        try:
            from yaml import full_load as yload
            state_str = yload(stream)
        except:
            import json
            state_str = json.load(stream)
    state = {}
    for k, s in state_str.items():
        if isinstance(s, str):
            state[k] = json.loads(s)
        else:
            state[k] = s
    return state


""" for load ASAPXYZ """


def load_asapxyz(data_spec):
    from asaplib.data import ASAPXYZ
    return ASAPXYZ(data_spec['fxyz'], data_spec['stride'], data_spec['periodic'], data_spec['fxyz_format'])


"""for gen_desc"""


def set_reducer(reducer_type, element_wise, zeta):
    """
    setting up the reducer function that is used to convert atomic descriptors into global descriptors for a structure.
    At the moment only one single reducer function can be used.
    """
    reducer_func = {'reducer1': {
        'reducer_type': reducer_type,  # [average], [sum], [moment_average], [moment_sum]
        'element_wise': element_wise,
    }}
    if reducer_type == 'moment_average' or reducer_type == 'moment_sum':
        reducer_func['zeta'] = zeta
    return reducer_func


def output_desc(asapxyz, desc_spec, prefix, peratom=False, N_processes=8):
    """
    Compute and save the descriptors
    """
    for tag, desc in desc_spec.items():
        # compute the descripitors
        asapxyz.compute_global_descriptors(desc_spec_dict=desc,
                                           sbs=[],
                                           keep_atomic=peratom,
                                           tag=tag,
                                           n_process=N_processes)
    asapxyz.write(prefix)
    asapxyz.save_state(prefix)


""" for maps and fits """


def read_xyz_n_dm(fxyz, design_matrix, use_atomic_descriptors, only_use_species, peratom):
    dm = []
    dm_atomic = []
    # try to read the xyz file
    if fxyz is not None and fxyz != 'none':
        from asaplib.data import ASAPXYZ
        asapxyz = ASAPXYZ(fxyz)
        if use_atomic_descriptors:
            dm = asapxyz.get_atomic_descriptors(design_matrix, only_use_species)
        else:
            dm, dm_atomic = asapxyz.get_descriptors(design_matrix, peratom)
    else:
        asapxyz = None
        print("Did not provide the xyz file. We can only output descriptor matrix.")
    # we can also load the descriptor matrix from a standalone file
    import os
    if os.path.isfile(design_matrix[0]):
        try:
            import numpy as np
            dm = np.genfromtxt(design_matrix[0], dtype=float)
            print("loaded the descriptor matrix from file: ", design_matrix[0])
        except:
            raise ValueError('Cannot load the descriptor matrix from file')
    return asapxyz, dm, dm_atomic


"""for maps"""


def figure_style_setups(prefix,
                        colorlabel, colorscale, colormap,
                        style, aspect_ratio, adjusttext):
    fig_options = {'outfile': prefix,
                   'show': False,
                   'title': None,
                   'size': [8 * aspect_ratio, 8],
                   'cmap': colormap,
                   'components': {
                       'first_p': {'type': 'scatter', 'clabel': colorlabel,
                                   'vmin': colorscale[0], 'vmax': colorscale[1]},
                       'second_p': {"type": 'annotate', 'adtext': adjusttext}}
                   }
    if style == 'journal':
        fig_options.update({'xlabel': None, 'ylabel': None,
                            'xaxis': False, 'yaxis': False,
                            'remove_tick': True,
                            'rasterized': True,
                            'fontsize': 12,
                            'size': [4 * aspect_ratio, 4]
                            })
    return fig_options


def map_process(obj, reduce_dict, axes, map_name):
    """
    process the dimensionality reduction command
    """
    # project
    if 'type' in reduce_dict.keys() and reduce_dict['type'] == 'RAW':
        proj = obj['design_matrix']
        if obj['map_options']['peratom']:
            proj_atomic = obj['design_matrix_atomic']
        else:
            proj_atomic = None
    else:
        from asaplib.reducedim import Dimension_Reducers
        dreducer = Dimension_Reducers(reduce_dict)
        proj = dreducer.fit_transform(obj['design_matrix'])
        if obj['map_options']['peratom']:
            print("Project atomic design matrix with No. of samples:", len(obj['design_matrix_atomic']))
            proj_atomic = dreducer.transform(obj['design_matrix_atomic'])
        else:
            proj_atomic = None
    # plot
    fig_spec = obj['fig_options']
    plotcolor = obj['map_options']['color']
    plotcolor_atomic = obj['map_options']['color_atomic']
    annotate = obj['map_options']['annotate']
    if 'cluster_labels' in obj.keys():
        labels = obj['cluster_labels']
    else:
        labels = []
    map_plot(fig_spec, proj, proj_atomic, plotcolor, plotcolor_atomic, labels, annotate, axes)
    # output 
    outfilename = obj['fig_options']['outfile']
    outmode = obj['map_options']['outmode']
    species_name = obj['map_options']['only_use_species']
    if obj['map_options']['project_atomic']:
        map_save(outfilename, outmode, obj['asapxyz'], None, proj, map_name, species_name)
    else:
        map_save(outfilename, outmode, obj['asapxyz'], proj, proj_atomic, map_name, species_name)


def map_plot(fig_spec, proj, proj_atomic, plotcolor, plotcolor_atomic, labels, annotate, axes):
    """
    Make plots
    """
    from matplotlib import pyplot as plt
    from asaplib.plot import Plotters
    asap_plot = Plotters(fig_spec)
    asap_plot.plot(proj[:, axes], plotcolor[:], labels[:], annotate[:])
    if proj_atomic is not None:
        asap_plot.plot(proj_atomic[:, axes], plotcolor_atomic[:], [], [])
    plt.show()


def map_save(foutput, outmode, asapxyz, proj, proj_atomic, map_name, species_name):
    """
    Save the low-D projections
    """
    if outmode == 'matrix':
        import numpy as np
        if proj is not None:
            np.savetxt(foutput + ".coord", proj, fmt='%4.8f', header='low D coordinates of samples')
        if proj_atomic is not None:
            np.savetxt(foutput + "-atomic.coord", proj_atomic, fmt='%4.8f', header=map_name)
    elif outmode in ('xyz', 'chemiscope'):
        if proj is not None:
            asapxyz.set_descriptors(proj, map_name)
        if proj_atomic is not None:
            asapxyz.set_atomic_descriptors(proj_atomic, map_name, species_name)
        if outmode == 'xyz':
            asapxyz.write(foutput)
        else:
            # If we write atomic projection assume we want to show them
            cutoff = 3.5 if proj_atomic else None
            asapxyz.write_chemiscope(foutput, cutoff=cutoff)
    else:
        pass


""" for clustering """


def cluster_process(asapxyz, trainer, design_matrix, cluster_options):
    """handle clustering operations"""
    prefix = cluster_options['prefix']

    from asaplib.cluster import DBCluster
    do_clustering = DBCluster(trainer)
    do_clustering.fit(design_matrix)

    do_clustering.save_state(prefix)

    labels_db = do_clustering.get_cluster_labels()
    if cluster_options['savexyz'] and asapxyz is not None:
        if cluster_options['use_atomic_descriptors']:
            asapxyz.set_atomic_descriptors(labels_db, prefix + '_cluster_label', cluster_options['only_use_species'])
        else:
            asapxyz.set_descriptors(labels_db, prefix + '_cluster_label')
        asapxyz.write(prefix)
    if cluster_options['savetxt']:
        import numpy as np
        np.savetxt(prefix + "-cluster-label.dat", np.transpose([np.arange(len(labels_db)), labels_db]),
                   header='index cluster_label', fmt='%d %d')

    return labels_db


""" for KDE """


def kde_process(asapxyz, density_model, proj, kde_options):
    """handle kernel density estimation operations"""
    # fit density model to data
    try:
        density_model.fit(proj)
    except:
        raise RuntimeError('KDE did not work. Try smaller dimension may help.')

    rho = density_model.evaluate_density(proj)
    prefix = kde_options['prefix']
    # save the density
    if kde_options['savetxt']:
        np.savetxt(prefix + "-kde.dat", np.transpose([np.arange(len(rho)), rho]),
                   header='index log_of_kernel_density_estimation', fmt='%d %4.8f')
    if kde_options['savexyz'] and asapxyz is not None:
        if kde_options['use_atomic_descriptors']:
            asapxyz.set_atomic_descriptors(rho, density_model.get_acronym(), kde_options['only_use_species'])
        else:
            asapxyz.set_descriptors(rho, density_model.get_acronym())
        asapxyz.write(prefix)
    return rho



"""
Adaptors for generating ChemiScope compatible inputs
"""

import warnings
import numpy as np
import json
import gzip

IGNORED_ASE_ARRAYS = ['positions', 'numbers', 'stress']


def _typetransform(data):
    """Ensure data is a list of JSON serialisable objects"""
    assert isinstance(data, list) and len(data) > 0
    if isinstance(data[0], str):
        return list(map(str, data))
    elif isinstance(data[0], bytes):
        return list(map(lambda u: u.decode('utf8'), data))
    else:
        try:
            if isinstance(data[0], float):
                return [float(value) for value in data]
            else:
                # 2D
                return [float(np.asarray(value).ravel()[0]) for value in data]
        except ValueError:
            raise Exception('unsupported type in value')


def _linearize(name, value):
    """
    Transform 2D arrays in multiple 1D arrays, converting types to fit json as
    needed.
    """
    data = {}
    if isinstance(value['values'], list):
        data[name] = {
            'target': value['target'],
            'values': _typetransform(value['values']),
        }
    elif isinstance(value['values'], np.ndarray):
        if len(value['values'].shape) == 1:
            data[name] = {
                'target': value['target'],
                'values': _typetransform(list(value['values'])),
            }
        elif len(value['values'].shape) == 2:
            for i in range(value['values'].shape[1]):
                data[f'{name}[{i + 1}]'] = {
                    'target': value['target'],
                    'values': _typetransform(list(value['values'][:, i])),
                }
        else:
            raise Exception('unsupported ndarray value')
    else:
        raise Exception(f'unknown type for value {name}')

    return data


def _frame_to_json(frame):
    data = {}
    data['size'] = len(frame)
    data['names'] = list(frame.symbols)
    data['x'] = [float(value) for value in frame.positions[:, 0]]
    data['y'] = [float(value) for value in frame.positions[:, 1]]
    data['z'] = [float(value) for value in frame.positions[:, 2]]

    if (frame.cell.lengths() != [0.0, 0.0, 0.0]).all():
        data['cell'] = list(np.concatenate(frame.cell))

    return data


def _generate_environments(frames, cutoff):
    environments = []
    for frame_id, frame in enumerate(frames):
        for center in range(len(frame)):
            environments.append({
                'structure': frame_id,
                'center': center,
                'cutoff': cutoff,
            })
    return environments


def write_chemiscope_input(filename,
                           frames,
                           meta=None,
                           extra=None,
                           cutoff=None):
    """
    Write the json file expected by the default chemiscope visualizer at
    ``filename``.
    :param str filename: name of the file to use to save the json data. If it
                         ends with '.gz', a gzip compressed file will be written
    :param list frames: list of `ase.Atoms`_ objects containing all the
                        structures
    :param dict meta: optional metadata of the dataset, see below
    :param dict extra: optional dictionary of additional properties, see below
    :param float cutoff: optional. If present, will be used to generate
                         atom-centered environments
    The dataset metadata should be given in the ``meta`` dictionary, the
    possible keys are:
    .. code-block:: python
        meta = {
            'name': '...',         # str, dataset name
            'description': '...',  # str, dataset description
            'authors': [           # list of str, dataset authors, OPTIONAL
                '...',
            ],
            'references': [        # list of str, references for this dataset,
                '...',             # OPTIONAL
            ],
        }
    The written JSON file will contain all the properties defined on the
    `ase.Atoms`_ objects. Values in ``ase.Atoms.arrays`` are mapped to
    ``target = "atom"`` properties; while values in ``ase.Atoms.info`` are
    mapped to ``target = "structure"`` properties. The only exception is
    ``ase.Atoms.arrays["numbers"]``, which is always ignored. If you want to
    have the atomic numbers as a property, you should add it to ``extra``
    manually.
    Additional properties can be added with the ``extra`` parameter. This
    parameter should be a dictionary containing one entry for each property.
    Each entry contains a ``target`` attribute (``'atom'`` or ``'structure'``)
    and a set of values. ``values`` can be a Python list of float or string; a
    1D numpy array of numeric values; or a 2D numpy array of numeric values. In
    the later case, multiple properties will be generated along the second axis.
    For example, passing
    .. code-block:: python
        extra = {
            'cheese': {
                'target': 'atom',
                'values': np.zeros((300, 4))
            }
        }
    will generate four properties named ``cheese[1]``, ``cheese[2]``,
    ``cheese[3]``,  and ``cheese[4]``, each containing 300 values.
    .. _`ase.Atoms`: https://wiki.fysik.dtu.dk/ase/ase/atoms.html
    
    :NOTE:
      Adapted from: https://github.com/cosmo-epfl/chemiscope/blob/master/utils/chemiscope_input.py
    """

    if not (filename.endswith('.json') or filename.endswith('.json.gz')):
        raise Exception('filename should end with .json or .json.gz')

    data = {'meta': {}}

    if meta is not None:
        if 'name' in meta:
            data['meta']['name'] = str(meta['name'])

        if 'description' in meta:
            data['meta']['description'] = str(meta['description'])

        if 'authors' in meta:
            data['meta']['authors'] = list(map(str, meta['authors']))

        if 'references' in meta:
            data['meta']['references'] = list(map(str, meta['references']))

        for key in meta.keys():
            if key not in ['name', 'description', 'authors', 'references']:
                warnings.warn('ignoring unexpected metadata: {}'.format(key))

    if 'name' not in data['meta'] or not data['meta']['name']:
        data['meta']['name'] = filename

    properties = {}
    if extra is not None:
        for name, value in extra.items():
            properties.update(_linearize(name, value))

    # Read properties coming from the ase.Atoms objects
    from_frames = {}

    # target: structure properties
    # TODO this need to updates as ASAP store arrays for each atom in the INFO?
    def _append_value(from_frames, name, value):
        """Append value to from_frames, create the entry if not exists"""
        if name in from_frames:
            from_frames[name]['values'].append(value)
        else:
            from_frames.update(
                {name: {
                    'target': 'structure',
                    'values': [value]
                }})
        return

    def _extend_value(from_frames, name, value):
        """Extend the entry in from_frames, create the entry if not exists"""
        if name in from_frames:
            from_frames[name]['values'].extend(value)
        else:
            from_frames.update(
                {name: {
                    'target': 'atom',
                    'values': list(value)
                }})
        return


    for frame in frames:
        for name, value in frame.info.items():
            if isinstance(value, (list, tuple, np.ndarray)):
                for idx, _value in enumerate(value):
                    _name = name + f'[{idx}]'
                    _append_value(from_frames, _name, _value)
            else:
                _append_value(from_frames, name, value)

    # target: atom properties
    has_atomic = False
    for frame in frames:
        for name, value in frame.arrays.items():
            if name in IGNORED_ASE_ARRAYS:
                continue
            has_atomic = True
            if len(value.shape) > 1:
                # Iterate over the columns
                for idx, _value in enumerate(value.T):
                    _name = 'atomic-' + name + f'[{idx}]'
                    _extend_value(from_frames, _name, _value)
            else:
                _extend_value(from_frames, name, value)

    for name, value in from_frames.items():
        properties.update(_linearize(name, value))

    data['properties'] = properties
    data['structures'] = [_frame_to_json(frame) for frame in frames]

    if cutoff is not None and has_atomic:
        data['environments'] = _generate_environments(frames, cutoff)

    if filename.endswith(".gz"):
        with gzip.open(filename, 'w', 9) as file:
            file.write(json.dumps(data).encode("utf8"))
    else:
        with open(filename, 'w') as file:
            json.dump(data, file)



"""
Tools for performing sparse kernel ridge regression of design matrices
using sparse KRR
NOTE: this class operate directly on design matrices.
for methods handingly kmatrix, see ./krr.py
"""

import numpy as np

from ..compressor import Sparsifier
from ..kernel import Descriptors_to_Kernels
from .base import RegressorBase

class SPARSE_KRR_Wrapper(RegressorBase):

    def __init__(self, kernel, krr_obj, sparse_mode="fps", n_sparse=None):
        """
        Parameters
        ----------
        kernel: dictionaries that specify which way to convert descriptors into kernel matrix
                     See ../kernel/kernel_transforms.py
        e.g.
        kernel = {
        "first_kernel": {"type": 'linear', "normalize": True}
        }
        e.g.
        {'k0':{"type": "cosine"}} 
        e.g.
        { 'k1': {"type": "polynomial", "d": power}}

        krr_obj: object for doing krr. Must have .fit() and .predict() methods
        sparsemode: str, default='fps', Sparsification method to use ([fps], [cur])'
        n_sparse: int, number of the representative samples, negative means no sparsification
         
        """
        self.sparse_mode = sparse_mode
        self.n_sparse = n_sparse
        if self.n_sparse is None or self.n_sparse > 0:
            self.sparsifier = Sparsifier(self.sparse_mode)
        else:
            self.sparsifier = None

        self.kernel = kernel
        # object for transform design matrix to kernel matrix
        self.k_transform = Descriptors_to_Kernels(kernel)

        # object for doing krr
        self.krr = krr_obj

        # we need to store the representative structures
        self.sbs = None
        # design matrix of the representative structures
        self.desc_sbs = None

        # state of the fit
        self._fitted = False

    def _sparsify(self, desc):
        # sparsification
        n_sample = len(desc)
        # set default value of n_sparse
        if self.n_sparse is None:
            self.n_sparse = max(10, n_sample // 20)
        # sparsification
        if self.n_sparse > 0 and self.n_sparse < n_sample:
            self.sbs = self.sparsifier.sparsify(desc, self.n_sparse)
        else:
            print("Not using any sparsification")
            self.sbs = range(n_sample)

        self.desc_sbs = desc[self.sbs]

    def fit(self, X, y):
        """
        Train the krr model with the design matrix and trainLabel.
        Parameters
        ----------
        X : array-like, shape=[n_descriptors, n_samples]
        Input points.
        y : array-like, shape=[n_samples]
        Input points.
        """
        print("a total of ", np.shape(X), "column")

        #if self._fitted:
        #    raise RuntimeError('SPARSE_KRR already fitted before, please reinitialise the object!')

        '''N train structures, M sparsified representative structures '''
        '''kMM: the kernel matrix of the representative structures with shape (M,M)'''
        '''kNM: the kernel matrix between the representative and the train structures with shape (N,M)'''
        self._sparsify(X)
        kMM = self.k_transform.compute(self.desc_sbs)
        kNM = self.k_transform.compute(X, self.desc_sbs)
        print(np.shape(kNM))
        self.krr.fit(kMM, y, kNM)

        self._fitted = True

    def predict(self, X):
        '''desc.shape is expected as [n_descriptors, n_samples]'''
        if not self._fitted:
            raise RuntimeError("The model has not been fitted yet, please fit it and then use predict.")
        kNM = self.k_transform.compute(X, self.desc_sbs)
        return self.krr.predict(kNM)

    def get_params(self, deep=True):
        pass

    def set_params(self, params, deep=True):
        pass

    def pack(self):
        pass

    def unpack(self, state):
        pass

    def loads(self, state):
        pass



"""
Tools to analyze clustering results
"""

from collections import Counter

import numpy as np


def output_cluster(prefix, labels, dicttags, tags):
    ofile = open("clustered-" + prefix + ".txt", "w")
    # output
    ofile.write("#! tag cluster_id \n")
    ndict = len(dicttags)
    ndata = len(tags)
    for i in range(ndict):
        ofile.write("%s %d\n" % (dicttags[i], labels[i]))
    for i in range(ndata):
        ofile.write("%s %d\n" % (tags[i], labels[i + ndict]))
    ofile.close()

    return 0


def output_cluster_sort(prefix, labels, dicttags, tags):
    # we sort the clusters as well
    ofile = open("sorted-clustered-" + prefix + ".txt", "w")
    # output
    ofile.write("#! tag cluster_id \n")
    ndict = len(dicttags)
    ndata = len(tags)

    sortlabels = np.stack((range(len(labels)), labels), axis=-1)
    sortlabels = sortlabels[sortlabels[:, 1].argsort()]

    for i, l in sortlabels:
        # print i,l
        if l >= 0 and i < ndict:
            ofile.write("%d %s\n" % (l, dicttags[i]))
        elif l >= 0:
            ofile.write("%d %s\n" % (l, tags[i - ndict]))
    return 0


def get_cluster_size(labels):
    unique_labels = set(labels)
    count = Counter(labels)
    return unique_labels, count


def most_frequent(List):
    occurence_count = Counter(List)
    return occurence_count.most_common(1)[0][0]


def array_handling(plist, attribute='mean'):
    """ available attributes:
        mean, sum, min, max, mode, all
    """
    if attribute == 'mean':
        return np.mean(plist)
    elif attribute == 'sum':
        return np.sum(plist)
    elif attribute == 'min':
        return np.amin(plist)
    elif attribute == 'max':
        return np.amax(plist)
    elif attribute == 'mode':
        return most_frequent(plist)
    elif attribute == 'all':
        return plist
    else:
        raise NameError('Attribute not found.')


def get_cluster_properties(labels, properties, attribute='mean'):
    unique_labels = set(labels)

    sortlabels = np.stack((range(len(labels)), labels), axis=-1)
    sortlabels = sortlabels[sortlabels[:, 1].argsort()]
    propertiesdict = {-1: 'noise'}

    ol = -1
    n = 0
    plist = []
    for i, l in sortlabels:
        if l > ol and l >= 0:
            propertiesdict[ol] = array_handling(plist, attribute)
            plist = []
        plist.append(properties[i])
        ol = l
    propertiesdict[ol] = array_handling(plist, attribute)

    return unique_labels, propertiesdict


def get_cluster_weighted_avg_properties(labels, properties, weights):
    unique_labels = set(labels)

    sortlabels = np.stack((range(len(labels)), labels), axis=-1)
    sortlabels = sortlabels[sortlabels[:, 1].argsort()]
    propertiesdict = {-1: 'noise'}

    ol = -1
    n = 0
    plist = []
    wlist = []
    for i, l in sortlabels:
        if l > ol and l >= 0:
            propertiesdict[ol] = np.mean(plist) / np.mean(wlist)
            plist = []
            wlist = []
        plist.append(properties[i] * weights[i])
        wlist.append(weights[i])
        ol = l
    propertiesdict[ol] = np.mean(plist) / np.mean(wlist)

    return unique_labels, propertiesdict



"""
Module containing the top level asap command
"""
# import numpy as np
import warnings

warnings.filterwarnings("ignore")

from .func_asap import *
from .cmd_cli_options import *
from asaplib.io import ConvertStrToList


@click.group('asap')
@click.pass_context
def asap(ctx):
    """The command line tool of Automatic Selection And Prediction tools for materials and molecules (ASAP)"""
    # Configure, make sure we have a dict object
    ctx.ensure_object(dict)
    """ stores the input data, with keys ['fxyz'], ['stride'], [periodic] """
    ctx.obj['data'] = {}
    """ stores the ASAPXYZ object """
    ctx.obj['asapxyz'] = None
    """ stores a np matrix used as the design matrix """
    ctx.obj['designmatrix'] = None
    """ stores a Design_Matrix object """
    ctx.obj['dm'] = None
    """ stores the specifications of descriptors. i.e. {tag: {desc_spec}} """
    ctx.obj['descriptors'] = {}
    """stores the specifications of generating descriptor options"""
    ctx.obj['desc_options'] = {}
    """stores the specifications of clustering options"""
    ctx.obj['cluster_options'] = {}
    """stores the specifications of kde options"""
    ctx.obj['kde_options'] = {}
    """stores the specifications of fit options"""
    ctx.obj['fit_options'] = {}
    """stores the specifications of dimensionality reduction map"""
    ctx.obj['map_options'] = {}
    """stores the specifications of the output figure"""
    ctx.obj['fig_options'] = {}


@asap.command('select')
@click.option('--algorithm', '--algo', '-a',
              type=click.Choice(['random', 'fps', 'cur'], case_sensitive=False),
              help='Sparsification algorithm to use',
              show_default=True, default='fps')
@click.option('--nkeep', '-n', type=int,
              help='The number (int) or the ratio (float) of samples to keep.',
              show_default=False, default=1)
@click.option('--design_matrix', '-dm', cls=ConvertStrToList, default='[]',
              help='Location of descriptor matrix file or name of the tags in ase xyz file\
                 the type is a list  \'[dm1, dm2]\', \
                as we can put together simutanously several design matrix.')
@click.pass_context
@file_input_options
@file_output_options
@output_setup_options
def select(ctx, fxyz, design_matrix,
           algorithm, nkeep,
           prefix, savexyz, savetxt):
    """
    Select a subset of frames using sparsification algorithms
    """

    if not fxyz and not design_matrix[0]:
        return

    if prefix is None:
        prefix = "ASAP-select-" + algorithm + "-n-" + str(nkeep)
    ctx.obj['asapxyz'], ctx.obj['design_matrix'], _ = read_xyz_n_dm(fxyz, design_matrix, False, False, False)

    from asaplib.compressor import Sparsifier
    sparsifier = Sparsifier(algorithm)
    sbs = sparsifier.sparsify(ctx.obj['design_matrix'], nkeep)
    # save
    if savetxt:
        selection = np.zeros(asapxyz.get_num_frames(), dtype=int)
        for i in sbs:
            selection[i] = 1
        np.savetxt(prefix + '.index', selection, fmt='%d')
    if savexyz: ctx.obj['asapxyz'].write(prefix, sbs)


@asap.group('gen_desc')
@click.option('--stride', '-s',
              help='Read in the xyz trajectory with X stide. Default: read/compute all frames.',
              default=1)
@click.option('--periodic/--no-periodic',
              help='Is the system periodic? If not specified, will infer from the XYZ file.',
              default=True)
@click.pass_context
@state_input_options
@file_input_options
@file_input_format_options
@file_output_options
@para_options
def gen_desc(ctx, in_file, fxyz, fxyz_format, prefix, stride, periodic, number_processes):
    """
    Descriptor generation command
    This command function evaluated before the descriptor specific ones,
    we setup the general stuff here, such as read the files.
    """

    if not fxyz and not in_file:
        return

    if in_file:
        # Here goes the routine to compute the descriptors according to the
        # state file(s)
        ctx.obj.update(load_in_file(in_file))

    if prefix is None: prefix = "ASAP-desc"
    if fxyz is not None:
        ctx.obj['data']['fxyz'] = fxyz
        ctx.obj['data']['fxyz_format'] = fxyz_format
        ctx.obj['data']['stride'] = stride
        ctx.obj['data']['periodic'] = periodic
    ctx.obj['desc_options']['prefix'] = prefix
    ctx.obj['desc_options']['N_processes'] = number_processes


@gen_desc.command('soap')
@click.option('--cutoff', '-c', type=float,
              help='Cutoff radius',
              show_default=False, default=None)
@click.option('--nmax', '-n', type=int,
              help='Maximum radial label',
              show_default=False, default=None)
@click.option('--lmax', '-l', type=int,
              help='Maximum angular label (<= 9)',
              show_default=False, default=None)
@click.option('--rbf', type=click.Choice(['gto', 'polynomial'], case_sensitive=False),
              help='Radial basis function',
              show_default=True, default='gto')
@click.option('--atom-gaussian-width', '-sigma', '-g', type=float,
              help='The width of the Gaussian centered on atoms.',
              show_default=True, default=0.5)
@click.option('--crossover/--no-crossover',
              help='If to included the crossover of atomic types.',
              show_default=True, default=False)
@click.option('--universal_soap', '--usoap', '-u',
              type=click.Choice(['none', 'smart', 'minimal', 'longrange'], case_sensitive=False),
              help='Try out our universal SOAP parameters.',
              show_default=True, default='minimal')
@click.pass_context
@desc_options
@atomic_to_global_desc_options
def soap(ctx, tag, cutoff, nmax, lmax, atom_gaussian_width, crossover, rbf, universal_soap,
         reducer_type, zeta, element_wise, peratom):
    """Generate SOAP descriptors"""
    # load up the xyz
    ctx.obj['asapxyz'] = load_asapxyz(ctx.obj['data'])

    if crossover is False:
        print("Warning: atomic species cross terms are not included! use --crossover if you want cross terms.")

    if universal_soap != 'none' and cutoff is None and nmax is None and lmax is None:
        from asaplib.hypers import universal_soap_hyper
        global_species = ctx.obj['asapxyz'].get_global_species()
        soap_spec = universal_soap_hyper(global_species, universal_soap, dump=True)
    elif cutoff is not None and nmax is not None and lmax is not None:
        soap_spec = {'soap1': {'type': 'SOAP',
                               'cutoff': cutoff,
                               'n': nmax,
                               'l': lmax,
                               'atom_gaussian_width': atom_gaussian_width}}
    else:
        raise IOError("Please either use universal soap or specify the values of cutoff, n, and l.")

    for k in soap_spec.keys():
        soap_spec[k]['rbf'] = rbf
        soap_spec[k]['crossover'] = crossover
    # The specification for the reducers
    reducer_spec = dict(set_reducer(reducer_type, element_wise, zeta))
    # The specification for the descriptor
    desc_spec = {}
    for k, v in soap_spec.items():
        desc_spec[k] = {'atomic_descriptor': dict({k: v}),
                        'reducer_function': reducer_spec}
    # specify descriptors using the cmd line tool
    ctx.obj['descriptors'][tag] = desc_spec
    # Compute the save the descriptors
    output_desc(ctx.obj['asapxyz'], ctx.obj['descriptors'], ctx.obj['desc_options']['prefix'], peratom,
                ctx.obj['desc_options']['N_processes'])


@gen_desc.command('acsf')
@click.option('--cutoff', '-c', type=float,
              help='Cutoff radius',
              show_default=False, default=None)
@click.option('--universal_acsf', '--uacsf', '-u',
              type=click.Choice(['none', 'smart', 'minimal', 'longrange'], case_sensitive=False),
              help='Try out our universal ACSF parameters.',
              show_default=True, default='minimal')
@click.pass_context
@desc_options
@atomic_to_global_desc_options
def acsf(ctx, tag, cutoff, universal_acsf,
         reducer_type, zeta, element_wise, peratom):
    """Generate ACSF descriptors"""
    # load up the xyz
    ctx.obj['asapxyz'] = load_asapxyz(ctx.obj['data'])

    from asaplib.hypers import universal_acsf_hyper
    global_species = ctx.obj['asapxyz'].get_global_species()
    if cutoff is not None:
        acsf_spec = universal_acsf_hyper(global_species, cutoff, dump=True, verbose=False)
    else:
        acsf_spec = universal_acsf_hyper(global_species, universal_acsf, dump=True, verbose=False)

    # The specification for the reducers
    reducer_spec = dict(set_reducer(reducer_type, element_wise, zeta))
    # The specification for the descriptor
    desc_spec = {'acsf': {'atomic_descriptor': acsf_spec,
                          'reducer_function': reducer_spec}
                 }
    # specify descriptors using the cmd line tool
    ctx.obj['descriptors'][tag] = desc_spec
    # Compute the save the descriptors
    output_desc(ctx.obj['asapxyz'], ctx.obj['descriptors'], ctx.obj['desc_options']['prefix'], peratom,
                ctx.obj['desc_options']['N_processes'])


@gen_desc.command('cm')
@click.pass_context
@desc_options
def cm(ctx, tag):
    """Generate the Coulomb Matrix descriptors"""
    # load up the xyz
    ctx.obj['asapxyz'] = load_asapxyz(ctx.obj['data'])
    # The specification for the descriptor
    ctx.obj['descriptors'][tag] = {'cm': {'type': "CM"}}
    # Compute the save the descriptors
    output_desc(ctx.obj['asapxyz'], ctx.obj['descriptors'], ctx.obj['desc_options']['prefix'], False,
                ctx.obj['desc_options']['N_processes'])


@gen_desc.command('run')
@click.pass_context
def run(ctx):
    """ Running analysis using input files """
    # load up the xyz
    ctx.obj['asapxyz'] = load_asapxyz(ctx.obj['data'])
    # Compute the save the descriptors
    output_desc(ctx.obj['asapxyz'], ctx.obj['descriptors'], ctx.obj['desc_options']['prefix'])


@asap.group('cluster', chain=True)
@click.pass_context
@file_input_options
@file_output_options
@dm_input_options
@km_input_options
@output_setup_options
def cluster(ctx, fxyz, design_matrix, use_atomic_descriptors, only_use_species, kernel_matrix,
            prefix, savexyz, savetxt):
    """
    Clustering using the design matrix.
    This command function evaluated before the specific ones,
    we setup the general stuff here, such as read the files.
    """

    if not fxyz and not design_matrix[0]:
        return
    if prefix is None: prefix = "ASAP-cluster"

    ctx.obj['cluster_options'] = {'prefix': prefix,
                                  'savexyz': savexyz,
                                  'savetxt': savetxt,
                                  'use_atomic_descriptors': use_atomic_descriptors,
                                  'only_use_species': only_use_species
                                  }

    ctx.obj['asapxyz'], ctx.obj['design_matrix'], _ = read_xyz_n_dm(fxyz, design_matrix, use_atomic_descriptors,
                                                                    only_use_species, False)

    if kernel_matrix != 'none':
        try:
            import numpy as np
            kNN = np.genfromtxt(kernel_matrix, dtype=float)
            print("loaded kernal matrix", kmat, "with shape", np.shape(kNN))
            from asaplib.kernel import kerneltodis
            ctx.obj['design_matrix'] = kerneltodis(kNN)
        except:
            raise ValueError('Cannot load the coordinates')


@cluster.command('fdb')
@click.pass_context
def fdb(ctx):
    """Clustering by fast search and find of density peaks (FDB)"""
    from asaplib.cluster import LAIO_DB
    trainer = LAIO_DB()

    ctx.obj['cluster_labels'] = cluster_process(ctx.obj['asapxyz'], trainer, ctx.obj['design_matrix'],
                                                ctx.obj['cluster_options'])


@cluster.command('dbscan')
@click.option('--metric', type=str,
              help='controls how distance is computed in the ambient space of the input data. \
                    See: https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html',
              show_default=True, default='euclidean')
@click.option('--min_samples', '-ms', type=int,
              help='The number of samples (or total weight) in a neighborhood for a point to be considered as a core point.',
              show_default=True, default=5)
@click.option('--eps', '-e', type=float,
              help='The maximum distance between two samples for one to be considered as in the neighborhood of the other.',
              default=None)
@click.pass_context
def dbscan(ctx, metric, min_samples, eps):
    """Density-based spatial clustering of applications with noise (DBSCAN)"""
    from asaplib.cluster import sklearn_DB
    if eps is None:
        from scipy.spatial.distance import cdist
        desc = ctx.obj['design_matrix']
        # we compute the characteristic bandwidth of the data
        # first select a subset of structures (20)
        import numpy as np
        sbs = np.random.choice(np.asarray(range(len(desc))), 50, replace=False)
        # the characteristic bandwidth of the data
        eps = np.percentile(cdist(desc[sbs], desc, metric), 100 * 10. / len(desc))

    trainer = sklearn_DB(eps, min_samples, metric)

    ctx.obj['cluster_labels'] = cluster_process(ctx.obj['asapxyz'], trainer, ctx.obj['design_matrix'],
                                                ctx.obj['cluster_options'])


@cluster.command('plot_pca')
@click.pass_context
@map_setup_options
@d_reduce_options
@file_output_options
def plot_pca(ctx, scale, dimension, axes,
             peratom, adjusttext, annotate, aspect_ratio, style, prefix):
    """ Plot the clustering results using a PCA map. Only use this command after fdb or dbscan.  """
    if prefix is None:
        prefix = "clustering-pca"
    colorlabel = "Clustering results"
    colorscale = [None, None]
    ctx.obj['map_options'].update({'color': ctx.obj['cluster_labels'],
                                   'color_atomic': [],
                                   'project_atomic': [],
                                   'only_use_species': ctx.obj['cluster_options']['only_use_species'],
                                   'peratom': False,
                                   'annotate': [],
                                   'outmode': 'none',
                                   'keepraw': True})
    if annotate != 'none':
        ctx.obj['map_options']['annotate'] = np.loadtxt(annotate, dtype="str")[:]

    ctx.obj['fig_options'] = figure_style_setups(prefix, colorlabel, colorscale, 'gnuplot', style, aspect_ratio,
                                                 adjusttext)
    ctx.obj['fig_options']['components'].update({"third_p": {"type": 'cluster', 'w_label': True, 'circle_size': 20}})
    map_name = "clustering-pca-d-" + str(dimension)
    reduce_dict = {'pca': {
        'type': 'PCA',
        'parameter': {"n_components": dimension, "scalecenter": scale}}
    }
    map_process(ctx.obj, reduce_dict, axes, map_name)


@asap.group('kde', chain=True)
@click.pass_context
@file_input_options
@file_output_options
@dm_input_options
@output_setup_options
def kde(ctx, fxyz, design_matrix, use_atomic_descriptors, only_use_species,
        prefix, savexyz, savetxt):
    """
    Kernel density estimation using the design matrix.
    This command function evaluated before the specific ones,
    we setup the general stuff here, such as read the files.
    """

    if not fxyz and design_matrix[0]:
        return
    if prefix is None: prefix = "ASAP-kde"

    ctx.obj['kde_options'] = {'prefix': prefix,
                              'savexyz': savexyz,
                              'savetxt': savetxt,
                              'use_atomic_descriptors': use_atomic_descriptors,
                              'only_use_species': only_use_species
                              }

    ctx.obj['asapxyz'], ctx.obj['design_matrix'], _ = read_xyz_n_dm(fxyz, design_matrix, use_atomic_descriptors,
                                                                    only_use_species, False)


@kde.command('kde_internal')
@click.option('--dimension', '-d', type=int,
              help='The number of the first D dimensions to keep when doing KDE.',
              show_default=True, default=8)
@click.pass_context
def kde_internal(ctx, dimension):
    """Internal implementation of KDE"""
    from asaplib.kde import KDE_internal
    import numpy as np
    proj = np.asmatrix(ctx.obj['design_matrix'])[:, 0:dimension]
    density_model = KDE_internal()
    ctx.obj['kde'] = kde_process(ctx.obj['asapxyz'], density_model, proj, ctx.obj['kde_options'])


@kde.command('kde_scipy')
@click.option('--dimension', '-d', type=int,
              help='The number of the first D dimensions to keep when doing KDE.',
              show_default=True, default=50)
@click.option('--bw_method', '-bw', type=str,
              help='This can be ‘scott’, ‘silverman’, a scalar constant or a callable.',
              show_default=False, default=None)
@click.pass_context
def kde_scipy(ctx, bw_method, dimension):
    """Scipy implementation of KDE"""
    from asaplib.kde import KDE_scipy
    import numpy as np
    proj = np.asmatrix(ctx.obj['design_matrix'])[:, 0:dimension]
    density_model = KDE_scipy(bw_method)
    ctx.obj['kde'] = kde_process(ctx.obj['asapxyz'], density_model, proj, ctx.obj['kde_options'])


@kde.command('kde_sklearn')
@click.option('--dimension', '-d', type=int,
              help='The number of the first D dimensions to keep when doing KDE.',
              show_default=True, default=50)
@click.option('--metric', type=str,
              help='controls how distance is computed in the ambient space of the input data. \
                    See: https://scikit-learn.org/stable/modules/density.html#kernel-density-estimation',
              show_default=True, default='euclidean')
@click.option('--algorithm', type=click.Choice(['kd_tree', 'ball_tree', 'auto'], case_sensitive=False),
              help='Algorithm to use',
              show_default=True, default='auto')
@click.option('--kernel', type=click.Choice(['gaussian', 'tophat', 'epanechnikov', 'exponential', 'linear', 'cosine'],
                                            case_sensitive=False),
              help='Kernel to use',
              show_default=True, default='gaussian')
@click.option('--bandwidth', '-bw', type=float,
              help='Bandwidth of the kernel',
              show_default=True, default=1)
@click.pass_context
def kde_scipy(ctx, dimension, bandwidth, algorithm, kernel, metric):
    """Scikit-learn implementation of KDE"""
    from asaplib.kde import KDE_sklearn
    import numpy as np
    proj = np.asmatrix(ctx.obj['design_matrix'])[:, 0:dimension]
    density_model = KDE_sklearn(bandwidth=bandwidth, algorithm=algorithm, kernel=kernel, metric=metric)
    ctx.obj['kde'] = kde_process(ctx.obj['asapxyz'], density_model, proj, ctx.obj['kde_options'])


@kde.command('plot_pca')
@click.pass_context
@map_setup_options
@d_reduce_options
@file_output_options
def plot_pca(ctx, scale, dimension, axes,
             peratom, adjusttext, annotate, aspect_ratio, style, prefix):
    """ Plot the KDE results using a PCA map """
    if prefix is None:
        prefix = "kde-pca"
    colorlabel = "Kernel density estimate (log scale)"
    colorscale = [None, None]
    ctx.obj['map_options'].update({'color': ctx.obj['kde'],
                                   'color_atomic': [],
                                   'project_atomic': [],
                                   'only_use_species': ctx.obj['kde_options']['only_use_species'],
                                   'peratom': False,
                                   'annotate': [],
                                   'outmode': 'none',
                                   'keepraw': True})
    if annotate != 'none':
        ctx.obj['map_options']['annotate'] = np.loadtxt(annotate, dtype="str")[:]

    ctx.obj['fig_options'] = figure_style_setups(prefix, colorlabel, colorscale, 'gnuplot', style, aspect_ratio,
                                                 adjusttext)
    map_name = "kde-pca-d-" + str(dimension)
    reduce_dict = {'pca': {
        'type': 'PCA',
        'parameter': {"n_components": dimension, "scalecenter": scale}}
    }
    map_process(ctx.obj, reduce_dict, axes, map_name)


@asap.group('map')
@click.pass_context
@file_input_options
@file_output_options
@dm_input_options
@map_setup_options
@map_io_options
@color_setup_options
def map(ctx, fxyz, design_matrix, prefix, output, extra_properties,
        use_atomic_descriptors, only_use_species, peratom, keepraw,
        color, color_column, color_label, colormap, color_from_zero, normalized_by_size,
        annotate, adjusttext, style, aspect_ratio):
    """
    Making 2D maps using dimensionality reduction.
    This command function evaluated before the specific ones,
    we setup the general stuff here, such as read the files.
    """

    if not fxyz and not design_matrix[0]:
        return
    if prefix is None: prefix = "ASAP-lowD-map"
    ctx.obj['asapxyz'], ctx.obj['design_matrix'], ctx.obj['design_matrix_atomic'] = read_xyz_n_dm(fxyz, design_matrix,
                                                                                                  use_atomic_descriptors,
                                                                                                  only_use_species,
                                                                                                  peratom)

    # Read additional properties
    if extra_properties:
        ctx.obj['asapxyz'].load_properties(extra_properties)

    if ctx.obj['asapxyz'] is None: output = 'matrix'
    print(len(ctx.obj['design_matrix_atomic']))
    # remove the raw descriptors
    if not keepraw and ctx.obj['asapxyz'] is not None:
        print("Remove raw desciptors..")
        ctx.obj['asapxyz'].remove_descriptors(design_matrix)
        ctx.obj['asapxyz'].remove_atomic_descriptors(design_matrix)

    # color scheme
    from asaplib.plot import set_color_function
    plotcolor, plotcolor_peratom, colorlabel, colorscale = set_color_function(color, ctx.obj['asapxyz'], color_column,
                                                                              0, peratom, use_atomic_descriptors,
                                                                              only_use_species, color_from_zero,
                                                                              normalized_by_size)
    if color_label is not None: colorlabel = color_label

    ctx.obj['map_options'] = {'color': plotcolor,
                              'color_atomic': plotcolor_peratom,
                              'project_atomic': use_atomic_descriptors,
                              'only_use_species': only_use_species,
                              'peratom': peratom,
                              'annotate': [],
                              'outmode': output,
                              'keepraw': keepraw
                              }
    if annotate != 'none':
        try:
            ctx.obj['map_options']['annotate'] = ctx.obj['asapxyz'].get_property(annotate)
        except:
            import numpy as np
            ctx.obj['map_options']['annotate'] = np.loadtxt(annotate, dtype="str")[:]

    ctx.obj['fig_options'] = figure_style_setups(prefix, colorlabel, colorscale, colormap, style, aspect_ratio,
                                                 adjusttext)


@map.command('raw')
@click.pass_context
@d_reduce_options
def raw(ctx, scale, dimension, axes):
    """Just plot the raw coordinates"""
    map_name = "raw-d-" + str(dimension)
    reduce_dict = {
        'type': 'RAW',
        'parameter': {"n_components": dimension, "scalecenter": scale}}
    map_process(ctx.obj, reduce_dict, axes, map_name)


@map.command('pca')
@click.pass_context
@d_reduce_options
def pca(ctx, scale, dimension, axes):
    """Principal Component Analysis"""
    map_name = "pca-d-" + str(dimension)
    reduce_dict = {'pca': {
        'type': 'PCA',
        'parameter': {"n_components": dimension, "scalecenter": scale}}
    }
    if scale:
        print("Perform standard scaling of the design matrix. To turn it off use `--no-scale`")
    map_process(ctx.obj, reduce_dict, axes, map_name)


@map.command('skpca')
@click.pass_context
@d_reduce_options
@kernel_options
@sparsification_options
def skpca(ctx, scale, dimension, axes,
          kernel, kernel_parameter, sparse_mode, n_sparse):
    """Sparse Kernel Principal Component Analysis"""
    map_name = "skpca-d-" + str(dimension)
    reduce_dict = {}
    if scale:
        print("Perform standard scaling of the design matrix. To turn it off use `--no-scale`")
        reduce_dict = {"preprocessing": {"type": 'SCALE', 'parameter': None}}
    reduce_dict['skpca'] = {"type": 'SPARSE_KPCA',
                            'parameter': {"n_components": dimension,
                                          "sparse_mode": sparse_mode, "n_sparse": n_sparse,
                                          "kernel": {"first_kernel": {"type": kernel, "d": kernel_parameter}}}}
    map_process(ctx.obj, reduce_dict, axes, map_name)


@map.command('umap')
@click.option('--n_neighbors', '-nn', type=int,
              help='Controls how UMAP balances local versus global structure in the data.',
              show_default=True, default=10)
@click.option('--min_dist', '-md', type=float,
              help='controls how tightly UMAP is allowed to pack points together.',
              show_default=True, default=0.1)
@click.option('--metric', type=str,
              help='controls how distance is computed in the ambient space of the input data. \
                    See: https://umap-learn.readthedocs.io/en/latest/parameters.html#metric',
              show_default=True, default='euclidean')
@click.pass_context
@d_reduce_options
def umap(ctx, scale, dimension, axes, n_neighbors, min_dist, metric):
    """UMAP"""
    map_name = "umap-d-" + str(dimension)
    reduce_dict = {}
    if scale:
        print("Perform standard scaling of the design matrix. To turn it off use `--no-scale`")
        reduce_dict = {"preprocessing": {"type": 'SCALE', 'parameter': None}}
    reduce_dict['umap'] = {'type': 'UMAP', 'parameter':
        {'n_components': dimension,
         'n_neighbors': n_neighbors,
         'min_dist': min_dist,
         'metric': metric
         }}
    map_process(ctx.obj, reduce_dict, axes, map_name)


@map.command('tsne')
@click.option('--metric', type=str,
              help='controls how distance is computed in the ambient space of the input data. \
                    See: https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html',
              show_default=True, default='euclidean')
@click.option('--learning_rate', '-l', type=float,
              help='The learning rate is usually in the range [10.0, 1000.0].',
              show_default=True, default=200.0)
@click.option('--early_exaggeration', '-e', type=float,
              help='Controls how tight natural clusters in the original space are in the embedded space.',
              show_default=True, default=12.0)
@click.option('--perplexity', '-p', type=float,
              help='The perplexity is related to the number of nearest neighbors that is used in other manifold learning algorithms. \
                  Larger datasets usually require a larger perplexity. \
               Consider selecting a value between 5 and 50. \
               Different values can result in significanlty different results.',
              show_default=True, default=30.0)
@click.option('--pca/--no-pca',
              help='Preprocessing the data using PCA with dimension 50. Recommended.',
              default=True)
@click.pass_context
@d_reduce_options
def tsne(ctx, pca, scale, dimension, axes,
         perplexity, early_exaggeration, learning_rate, metric):
    """t-SNE"""
    map_name = "tsne-d-" + str(dimension)
    reduce_dict = {}
    if pca:
        # pre-process with PCA if dim > 50
        # suggested here: https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
        reduce_dict = {"preprocessing": {"type": 'PCA', 'parameter': {"n_components": 50, "scalecenter": scale}}}
    elif scale:
        print("Perform standard scaling of the design matrix. To turn it off use `--no-scale`")
        reduce_dict = {"preprocessing": {"type": 'SCALE', 'parameter': None}}
    reduce_dict['tsne'] = {'type': 'TSNE', 'parameter':
        {'perplexity': perplexity,
         'early_exaggeration': early_exaggeration,
         'learning_rate': learning_rate,
         'metric': metric
         }}
    map_process(ctx.obj, reduce_dict, axes, map_name)


@asap.group('fit')
@click.pass_context
@file_input_options
@file_output_options
@dm_input_options
@fit_setup_options
def fit(ctx, fxyz, design_matrix, use_atomic_descriptors, only_use_species, y, normalized_by_size, prefix,
        test_ratio, learning_curve, lc_points):
    """
    Fit a machine learning model to the design matrix and labels.
    This command function evaluated before the specific ones,
    we setup the general stuff here, such as read the files.
    """
    if not fxyz and not design_matrix[0]:
        return
    if prefix is None: prefix = "ASAP-fit"

    ctx.obj['fit_options'] = {"prefix": prefix,
                              "learning_curve": learning_curve,
                              "lc_points": lc_points,
                              "test_ratio": test_ratio
                              }
    asapxyz, desc, _ = read_xyz_n_dm(fxyz, design_matrix, use_atomic_descriptors, only_use_species, False)

    try:
        import numpy as np
        y_all = np.genfromtxt(y, dtype=float)
    except:
        if use_atomic_descriptors:
            y_all = asapxyz.get_atomic_property(y, normalized_by_size)
        else:
            y_all = asapxyz.get_property(y, normalized_by_size)
    # print(y_all)

    from asaplib.data import Design_Matrix
    ctx.obj['dm'] = Design_Matrix(desc, y_all, True, test_ratio)


@fit.command('ridge')
@click.option('--sigma', type=float,
              help='the noise level of the signal. Also the regularizer that improves the stablity of matrix inversion.',
              default=0.0001)
@click.pass_context
def ridge(ctx, sigma):
    """Ridge Regression"""
    from asaplib.fit import RidgeRegression
    rr = RidgeRegression(sigma)
    # fit the model
    ctx.obj['dm'].compute_fit(rr, 'ridge_regression', store_results=True, plot=True)
    if ctx.obj['fit_options']["learning_curve"] > 1:
        ctx.obj['dm'].compute_learning_curve(rr, 'ridge_regression', ctx.obj['fit_options']["learning_curve"],
                                             ctx.obj['fit_options']["lc_points"], randomseed=42, verbose=False)

    ctx.obj['dm'].save_state(ctx.obj['fit_options']['prefix'])
    from matplotlib import pyplot as plt
    plt.show()


@fit.command('kernelridge')
@click.option('--sigma', type=float,
              help='the noise level of the signal. Also the regularizer that improves the stablity of matrix inversion.',
              default=0.0001)
@kernel_options
@sparsification_options
@click.pass_context
def kernelridge(ctx, sigma, kernel, kernel_parameter, sparse_mode, n_sparse):
    """Kernel Ridge Regression (with sparsification)"""
    from asaplib.fit import SPARSE_KRR_Wrapper, KRRSparse
    k_spec = {"first_kernel": {"type": kernel, "d": kernel_parameter}}
    krr = KRRSparse(0., None, sigma)
    skrr = SPARSE_KRR_Wrapper(k_spec, krr, sparse_mode=sparse_mode, n_sparse=n_sparse)
    # fit the model
    ctx.obj['dm'].compute_fit(skrr, 'skrr', store_results=True, plot=True)
    if ctx.obj['fit_options']["learning_curve"] > 1:
        ctx.obj['dm'].compute_learning_curve(skrr, 'skrr', ctx.obj['fit_options']["learning_curve"],
                                             ctx.obj['fit_options']["lc_points"], randomseed=42, verbose=False)

    ctx.obj['dm'].save_state(ctx.obj['fit_options']['prefix'])
    from matplotlib import pyplot as plt
    plt.show()



"""
Methods and functions to compute atomic desciptors
"""
import json

import numpy as np

from ..io import NpEncoder


class Atomic_Descriptors:
    def __init__(self, desc_spec_dict={}):
        """
        Object handing the specification and the computation of atomic descriptors
        Parameters
        ----------
        desc_spec_dict: dictionaries that specify which atomic descriptor to use 
        e.g.
        desc_spec_dict = {
        "firstsoap": 
        {"type": 'SOAP',"species": [1, 6, 7, 8], "cutoff": 2.0, "atom_gaussian_width": 0.2, "n": 4, "l": 4}
        }
        """
        self.desc_spec_dict = desc_spec_dict
        # list of Atomic_Descriptor objections
        self.engines = {}
        self.acronym = ""

        self.bind()

    def add(self, desc_spec, tag):
        """
        adding the specifications of a new atomic descriptors
        Parameters
        ----------
        desc_spec: a dictionary that specify which atomic descriptor to use 
        """
        self.desc_spec_dict[tag] = desc_spec

    def pack(self):
        return json.dumps(self.desc_spec_dict, sort_keys=True, cls=NpEncoder)

    def get_acronym(self):
        if self.acronym == "":
            for element in self.desc_spec_dict.keys():
                self.acronym += self.engines[element].get_acronym()
        return self.acronym

    def bind(self):
        """
        binds the objects that actually compute the descriptors
        these objects need to have .create(frame) method to compute the descriptors of frame (a xyz object)
        """
        # clear up the objects
        self.engines = {}
        for element in self.desc_spec_dict.keys():
            self.engines[element] = self._call(self.desc_spec_dict[element])
            self.desc_spec_dict[element]['acronym'] = self.engines[element].get_acronym()

    def _call(self, desc_spec):
        """
        call the specific descriptor objects
        """
        if "type" not in desc_spec.keys():
            raise ValueError("Did not specify the type of the descriptor.")
        if desc_spec["type"] == "SOAP":
            return Atomic_Descriptor_SOAP(desc_spec)
        if desc_spec["type"] == "ACSF":
            return Atomic_Descriptor_ACSF(desc_spec)
        if desc_spec["type"] == "LMBTR_K2":
            return Atomic_Descriptor_LMBTR_K2(desc_spec)
        if desc_spec["type"] == "LMBTR_K3":
            return Atomic_Descriptor_LMBTR_K3(desc_spec)
        if desc_spec["type"] == "FCHL19":
            return Atomic_Descriptor_FCHL19(desc_spec)
        else:
            raise NotImplementedError

    def compute(self, frame):
        """
        compute the global descriptor vector for a frame from atomic contributions
        Parameters
        ----------
        frame: ASE atom object. Coordinates of a frame.

        Returns
        -------
        atomic_desc_dict : a dictionary. each entry contains the essential info of the descriptor (acronym) 
                          and a np.array [N_desc*N_atoms]. Atomic descriptors for a frame.
                          e.g. {'ad1':{'acronym':'soap-1', 'atomic_descriptors': `a np.array [N_desc*N_atoms]`}}
        """
        atomic_desc_dict = {}
        for element in self.desc_spec_dict.keys():
            atomic_desc_dict[element] = {}
            atomic_desc_dict[element]['acronym'], atomic_desc_dict[element]['atomic_descriptors'] = self.engines[
                element].create(frame)
        return atomic_desc_dict


class Atomic_Descriptor_Base:
    def __init__(self, desc_spec):
        self._is_atomic = True
        self.acronym = ""
        pass

    def is_atomic(self):
        return self._is_atomic

    def get_acronym(self):
        """ 
        we use an acronym for each descriptor, so it's easy to find it and refer to it
        """
        return self.acronym

    def create(self, frame):
        """
        notice that we return the acronym here!!!
        """
        return self.acronym, []

    def _get_pbc(self, frame):
        return any(frame.get_pbc())

class Atomic_Descriptor_SOAP(Atomic_Descriptor_Base):
    def __init__(self, desc_spec):
        """
        make a DScribe SOAP object
        """

        from dscribe.descriptors import SOAP

        if "type" not in desc_spec.keys() or desc_spec["type"] != "SOAP":
            raise ValueError("Type is not SOAP or cannot find the type of the descriptor")

        # required
        try:
            self.species = desc_spec['species']
            self.cutoff = desc_spec['cutoff']
            self.g = desc_spec['atom_gaussian_width']
            self.n = desc_spec['n']
            self.l = desc_spec['l']
        except:
            raise ValueError("Not enough information to intialize the `Atomic_Descriptor_SOAP` object")

        # we have defaults here
        if 'rbf' in desc_spec.keys():
            self.rbf = desc_spec['rbf']
        else:
            self.rbf = 'gto'

        if 'crossover' in desc_spec.keys():
            self.crossover = bool(desc_spec['crossover'])
        else:
            self.crossover = False

        if 'periodic' in desc_spec.keys():
            self.periodic = bool(desc_spec['periodic'])
        else:
            self.periodic = True

        print("Using SOAP Descriptors ...")

        # make an acronym
        self.acronym = "SOAP-n" + str(self.n) + "-l" + str(self.l) + "-c" + str(self.cutoff) + "-g" + str(self.g)

    def create(self, frame):
        from dscribe.descriptors import SOAP
        self.soap = SOAP(species=self.species, rcut=self.cutoff, nmax=self.n, lmax=self.l,
                         sigma=self.g, rbf=self.rbf, crossover=self.crossover, average='off',
                         periodic=self._get_pbc(frame))

        # notice that we return the acronym here!!!
        return self.acronym, self.soap.create(frame, n_jobs=1)


class Atomic_Descriptor_ACSF(Atomic_Descriptor_Base):
    def __init__(self, desc_spec):
        """
        make a DScribe ACSF object

        see: 
        https://singroup.github.io/dscribe/tutorials/acsf.html

        # template for an ACSF descriptor
        # currenly Dscribe only supports ASCF for finite system!
        """

        if "type" not in desc_spec.keys() or desc_spec["type"] != "ACSF":
            raise ValueError("Type is not ACSF or cannot find the type of the descriptor")

        if 'periodic' in desc_spec.keys():
            self.periodic = bool(desc_spec['periodic'])
        if self.periodic == True:
            raise ValueError("Warning: currently DScribe only supports ACSF for finite systems")

        from dscribe.descriptors import ACSF

        self.acsf_dict = {
            'g2_params': None,
            'g3_params': None,
            'g4_params': None,
            'g5_params': None}

        # required
        try:
            self.species = desc_spec['species']
            self.cutoff = desc_spec['cutoff']
        except:
            raise ValueError("Not enough information to intialize the `Atomic_Descriptor_ACF` object")

        # fill in the values
        for k, v in desc_spec.items():
            if k in self.acsf_dict.keys():
                if isinstance(v, list):
                    self.acsf_dict[k] = np.asarray(v)
                else:
                    self.acsf_dict[k] = v

        self.acsf = ACSF(species=self.species, rcut=self.cutoff, **self.acsf_dict, sparse=False)

        print("Using ACSF Descriptors ...")

        # make an acronym
        self.acronym = "ACSF-c" + str(self.cutoff)
        if self.acsf_dict['g2_params'] is not None: self.acronym += "-g2-" + str(len(self.acsf_dict['g2_params']))
        if self.acsf_dict['g3_params'] is not None: self.acronym += "-g3-" + str(len(self.acsf_dict['g3_params']))
        if self.acsf_dict['g4_params'] is not None: self.acronym += "-g4-" + str(len(self.acsf_dict['g4_params']))
        if self.acsf_dict['g5_params'] is not None: self.acronym += "-g5-" + str(len(self.acsf_dict['g5_params']))

    def create(self, frame):
        # notice that we return the acronym here!!!
        return self.acronym, self.acsf.create(frame, n_jobs=1)


class Atomic_Descriptor_LMBTR(Atomic_Descriptor_Base):
    def __init__(self, desc_spec):
        """
        make a DScribe LMBTR object
        (see https://singroup.github.io/dscribe/tutorials/lmbtr.html)')
        Args:
            species:
            periodic (bool): Determines whether the system is considered to be
                periodic.
            k2 (dict): Dictionary containing the setup for the k=2 term.
                Contains setup for the used geometry function, discretization and
                weighting function. For example::

                    k2 = {
                        "geometry": {"function": "inverse_distance"},
                        "grid": {"min": 0.1, "max": 2, "sigma": 0.1, "n": 50},
                        "weighting": {"function": "exp", "scale": 0.75, "cutoff": 1e-2}
                    }

            k3 (dict): Dictionary containing the setup for the k=3 term.
                Contains setup for the used geometry function, discretization and
                weighting function. For example::

                    k3 = {
                        "geometry": {"function": "angle"},
                        "grid": {"min": 0, "max": 180, "sigma": 5, "n": 50},
                        "weighting" = {"function": "exp", "scale": 0.5, "cutoff": 1e-3}
                    }
            normalize_gaussians (bool): Determines whether the gaussians are
                normalized to an area of 1. Defaults to True. If False, the
                normalization factor is dropped and the gaussians have the form.
                :math:`e^{-(x-\mu)^2/2\sigma^2}`
            normalization (str): Determines the method for normalizing the
                output. The available options are:

                * "none": No normalization.
                * "l2_each": Normalize the Euclidean length of each k-term
                  individually to unity.

            flatten (bool): Whether the output should be flattened to a 1D
                array. If False, a dictionary of the different tensors is
                provided, containing the values under keys: "k1", "k2", and
                "k3":
            sparse (bool): Whether the output should be a sparse matrix or a
                dense numpy array.
        """

        # required
        try:
            self.species = desc_spec['species']
        except:
            raise ValueError("Not enough information to intialize the `Atomic_Descriptor_LMBTR` object")

        # we have defaults here
        if 'normalization' in desc_spec.keys():
            self.normalization = desc_spec['normalization']
        else:
            self.normalization = None  # or "l2_each"

        if 'normalize_gaussians' in desc_spec.keys():
            self.normalize_gaussians = desc_spec['normalize_gaussians']
        else:
            self.normalize_gaussians = "True"  # or False

        if 'periodic' in desc_spec.keys():
            self.periodic = bool(desc_spec['periodic'])
        else:
            self.periodic = True

    def create(self, frame):
        # notice that we return the acronym here!!!
        return self.acronym, self.lmbtr.create(frame, n_jobs=1)


class Atomic_Descriptor_LMBTR_K2(Atomic_Descriptor_LMBTR):
    def __init__(self, desc_spec):

        super().__init__(desc_spec)

        from dscribe.descriptors import LMBTR

        if "type" not in desc_spec.keys() or desc_spec["type"] != "LMBTR_K2":
            raise ValueError("Type is not LMBTR_K2 or cannot find the type of the descriptor")

        # required
        try:
            self.k2 = desc_spec['k2']
        except:
            raise ValueError("Not enough information to intialize the `Atomic_Descriptor_LMBTR` object")

        self.lmbtr = LMBTR(species=self.species, periodic=self.periodic, flatten=True,
                           normalize_gaussians=self.normalize_gaussians,
                           k2=self.k2)

        print("Using LMBTR-K2 Descriptors ...")

        # make an acronym
        self.acronym = "LMBTR-K2"  # perhaps add more info here


class Atomic_Descriptor_LMBTR_K3(Atomic_Descriptor_LMBTR):
    def __init__(self, desc_spec):

        super().__init__(desc_spec)

        from dscribe.descriptors import LMBTR

        if "type" not in desc_spec.keys() or desc_spec["type"] != "LMBTR_K3":
            raise ValueError("Type is not LMBTR_K3 or cannot find the type of the descriptor")

        # required
        try:
            self.k2 = desc_spec['k3']
        except:
            raise ValueError("Not enough information to intialize the `Atomic_Descriptor_LMBTR` object")

        self.lmbtr = LMBTR(species=self.species, periodic=self.periodic, flatten=True,
                           normalize_gaussians=self.normalize_gaussians,
                           k3=self.k3)

        print("Using LMBTR-K3 Descriptors ...")

        # make an acronym
        self.acronym = "LMBTR-K3"  # perhaps add more info here


class Atomic_Descriptor_FCHL19(Atomic_Descriptor_Base):
    """
    Generate the FCHL19 representation (https://doi.org/10.1063/1.5126701).
    Requires the developer version of the QML package, see
    https://www.qmlcode.org/installation.html for installation instructions.


    Parameters
    ----------
    :param nRs2: Number of gaussian basis functions in the two-body terms
    :type nRs2: integer
    :param nRs3: Number of gaussian basis functions in the three-body radial part
    :type nRs3: integer
    :param nFourier: Order of Fourier expansion
    :type nFourier: integer
    :param eta2: Precision in the gaussian basis functions in the two-body terms
    :type eta2: float
    :param eta3: Precision in the gaussian basis functions in the three-body radial part
    :type eta3: float
    :param zeta: Precision parameter of basis functions in the three-body angular part
    :type zeta: float
    :param two_body_decay: exponential decay for the two body function
    :type two_body_decay: float
    :param three_body_decay: exponential decay for the three body function
    :type three_body_decay: float
    :param three_body_weight: relative weight of the three body function
    :type three_body_weight: float
    :is_periodic: Boolean determining Whether the system is periodic.
    :type Boolean:
    """

    def __init__(self, desc_spec):

        if "type" not in desc_spec.keys() or desc_spec["type"] != "FCHL19":
            raise ValueError("Type is not FCHL19 or cannot find the type of the descriptor")

        if 'periodic' in desc_spec.keys():
            self.periodic = bool(desc_spec['periodic'])
        if self.periodic == True:
            raise ValueError("Warning: currently DScribe only supports FCHL19 for finite systems")

        print("Warning: This FCHL19 atomic descriptor is untested, because I (Bingqing) cannot install QML!")
        raise NotImplementedError

        self.fchl_acsf_dict = {'nRs2': None,
                               'nRs3': None,
                               'nFourier': None,
                               'eta2': None,
                               'eta3': None,
                               'zeta': None,
                               'rcut': None,
                               'acut': None,
                               'two_body_decay': None,
                               'three_body_decay': None,
                               'three_body_weight': None,
                               'pad': False, 'gradients': False}

        # required
        try:
            self.species = desc_spec['species']
        except:
            raise ValueError("Not enough information to intialize the `Atomic_Descriptor_ACF` object")

        # fill in the values
        for k, v in desc_spec.items():
            if k in self.fchl_acsf_dict.keys():
                self.fchl_acsf_dict[k] = v

        print("Using FCHL19 Descriptors ...")

        # make an acronym
        self.acronym = "FCHL19-c"  # add more stuff here

    def _repr_wrapper(frame, elements,
                      nRs2=24, nRs3=20,
                      nFourier=1, eta2=0.32, eta3=2.7,
                      zeta=np.pi, rcut=8.0, acut=8.0,
                      two_body_decay=1.8, three_body_decay=0.57,
                      three_body_weight=13.4, stride=1):

        nuclear_charges, coordinates = frame.get_atomic_numbers(), frame.get_positions()
        rep = generate_fchl_acsf(nuclear_charges, coordinates, elements,
                                 nRs2=nRs2, nRs3=nRs3, nFourier=nFourier,
                                 eta2=eta2, eta3=eta3, zeta=zeta,
                                 rcut=rcut, acut=acut,
                                 two_body_decay=two_body_decay, three_body_decay=three_body_decay,
                                 three_body_weight=three_body_weight,
                                 pad=False, gradients=False)
        rep_out = np.zeros((rep.shape[0], len(elements), rep.shape[1]))

        for i, z in enumerate(nuclear_charges):
            j = np.where(np.equal(z, elements))[0][0]
            rep_out[i, j] = rep[i]
        rep_out = rep_out.reshape(len(rep_out), -1)
        return rep_out

    def create(self, frame):
        # notice that we return the acronym here!!!
        return self.acronym, self_repr_wrapper(frame, self.species, **self.fchl_acsf_dict)



"""
Functions for assessing the quality of the fits
"""

import numpy as np
from scipy.stats import spearmanr
from scipy.stats import pearsonr

class LC_SCOREBOARD():
    def __init__(self, train_sizes):
        self.scores = {size: [] for size in train_sizes}

    def add_score(self, Ntrain, score):
        if Ntrain in self.scores:
            self.scores[Ntrain].append(score)
        else:
            self.scores[Ntrain] = [score]
    def dump_all(self):
        return self.scores

    def fetch_all(self):
        lc_results = {}
        for sc in score_func.keys():
            lc_results[sc] = self.fetch(sc)
        return lc_results

    def fetch(self, sc_name='RMSE'):
        Ntrains = []
        avg_scores = []
        avg_scores_error = []
        for Ntrain, score in self.scores.items():
            avg = 0.
            var = 0.
            for sc in score:
                avg += sc[sc_name]
                var += sc[sc_name] ** 2.
            avg /= len(score)
            var /= len(score)
            var -= avg ** 2.
            avg_scores.append(avg)
            avg_scores_error.append(np.sqrt(var))
            Ntrains.append(int(Ntrain))
        return np.stack((Ntrains, avg_scores, avg_scores_error), axis=-1).tolist()

    def plot_learning_curve(self, sc_name='RMSE'):
        """plot the learning curve"""
        from matplotlib import pyplot as plt
        lc_results = self.fetch(sc_name)
       
        fig, ax = plt.subplots()
        ax.errorbar(lc_results[:,0], lc_results[:,1], yerr=lc_results[:,2], linestyle='-', uplims=True, lolims=True)
        ax.set_title('Learning curve')
        ax.set_xlabel('Number of training samples')
        ax.set_ylabel('Test {}'.format(sc_name))
        ax.set_xscale('log')
        ax.set_yscale('log')
        return fig, ax

def get_score(ypred, y):
    scores = {}
    for k, func in score_func.items():
        scores[k] = func(ypred, y)
    return scores

def get_r2(y_pred, y):
    weight = 1
    sample_weight = None
    numerator = (weight * (y - y_pred) ** 2).sum(axis=0, dtype=np.float64)
    denominator = (weight * (y - np.average(
        y, axis=0, weights=sample_weight)) ** 2).sum(axis=0, dtype=np.float64)
    output_scores = 1 - (numerator / denominator)
    return np.mean(output_scores).tolist()


def get_mae(ypred, y):
    return np.mean(np.abs(ypred - y)).tolist()


def get_rmse(ypred, y):
    return np.sqrt(np.mean((ypred - y) ** 2)).tolist()


def get_sup(ypred, y):
    return np.amax(np.abs((ypred - y))).tolist()


def get_spearman(ypred, y):
    corr, _ = spearmanr(ypred, y)
    return corr

def get_pearson(ypred, y):
    corr, _ = pearsonr(ypred, y)
    return corr

score_func = dict(
    MAE=get_mae,
    RMSE=get_rmse,
    SUP=get_sup,
    R2=get_r2,
    SpearmanR=get_spearman,
    PearsonR=get_pearson
)









"""
I adapted the code from:
https://github.com/alexandreday/fast_density_clustering.git
Copyright 2017 Alexandre Day
"""

import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KernelDensity, NearestNeighbors

from .density_estimation import Kernel_Density_Base

class KDE_internal(Kernel_Density_Base):
    """Kernel density estimation (KDE) for accurate local density estimation.
    This is achieved by using maximum-likelihood estimation of the generative kernel density model
    which is regularized using cross-validation.


    Parameters
    ----------
    bandwidth: float, optional
        bandwidth for the kernel density estimation. If not specified, will be determined automatically using
        maximum likelihood on a test-set.

    nh_size : int, optional (default = 'auto')
        number of points in a typical neighborhood... only relevant for evaluating
        a crude estimate of the bandwidth.'auto' means that the nh_size is scaled with number of samples. We 
        use nh_size = 100 for 10000 samples. The minimum neighborhood size is set to 4.

    test_ratio_size: float, optional (default = 0.8)
        Ratio size of the test set used when performing maximum likehood estimation.
        In order to have smooth density estimations (prevent overfitting), it is recommended to
        use a large test_ratio_size (closer to 1.0) rather than a small one.

    atol: float, optional (default = 0.000005)
        kernel density estimate precision parameter. determines the precision used for kde.
        smaller values leads to slower execution but better precision
    
    rtol: float, optional (default = 0.00005)
        kernel density estimate precision parameter. determines the precision used for kde.
        smaller values leads to slower execution but better precision
    
    xtol: float, optional (default = 0.01)
        precision parameter for optimizing the bandwidth using maximum likelihood on a test set

    test_ratio_size: float, optional
        ratio of the test size for determining the bandwidth.

    kernel: str, optional (default='gaussian')
        Type of Kernel to use for density estimates. Other options are {'epanechnikov'|'linear','tophat'}.
    """

    def __init__(self, nh_size='auto', bandwidth=None, test_ratio_size=0.1,
                 xtol=0.01, atol=0.000005, rtol=0.00005, extreme_dist=False,
                 nn_dist=None, kernel='gaussian'):

        self.bandwidth = bandwidth
        self.nh_size = nh_size
        self.test_ratio_size = test_ratio_size
        self.xtol = xtol
        self.atol = atol
        self.rtol = rtol
        self.extreme_dist = extreme_dist
        self.nn_dist = nn_dist
        self.kernel = kernel  # epanechnikov other option

        self.acronym = 'kde_internal'

    def fit(self, X):

        """Fit kernel model to X"""

        if self.nh_size is 'auto':
            self.nh_size = max([int(25 * np.log10(X.shape[0])), 4])

        if X.shape[1] > 8:
            print('Careful, you are trying to do density estimation for data in a D > 8 dimensional space\n'
                  ' ... you are warned !')

        if self.bandwidth is None:
            self.bandwidth = self.find_optimal_bandwidth(X)
        else:
            self.kde = KernelDensity(
                bandwidth=self.bandwidth, algorithm='kd_tree',
                kernel=self.kernel, metric='euclidean',
                atol=self.atol, rtol=self.rtol,
                breadth_first=True, leaf_size=40
            )
        self.kde.fit(X)
        return self

    def evaluate_density(self, X):
        """Given an array of data, computes the local density of every point using kernel density estimation

        Input
        ------
        Data X : array, shape(n_sample,n_feature)

        Return
        ------
        Log of densities for every point: array, shape(n_sample)
        Return:
            kde.score_samples(X)
        """
        return self.kde.score_samples(X)

    def bandwidth_estimate(self, X_train, X_test):
        """Gives a rough estimate of the optimal bandwidth (based on the notion of some effective neigborhood)
        
        Return
        ---------
        bandwidth estimate, minimum possible value : tuple, shape(2)
        """

        if self.nn_dist is None:
            nn = NearestNeighbors(n_neighbors=self.nh_size, algorithm='kd_tree')
            nn.fit(X_train)
            nn_dist, _ = nn.kneighbors(X_test, n_neighbors=self.nh_size, return_distance=True)
        else:
            nn_dist = self.nn_dist

        dim = X_train.shape[1]

        # Computation of minimum bound
        # This can be computed by taking the limit h -> 0 and making a saddle-point approx.
        mean_nn2_dist = np.mean(nn_dist[:, 1] * nn_dist[:, 1])
        h_min = np.sqrt(mean_nn2_dist / dim)

        idx_1 = np.random.choice(np.arange(len(X_train)), size=min([1000, len(X_train)]), replace=False)
        idx_2 = np.random.choice(np.arange(len(X_test)), size=min([1000, len(X_test)]), replace=False)

        max_size = min([len(idx_1), len(idx_2)])

        tmp = np.linalg.norm(X_train[idx_1[:max_size]] - X_test[idx_2[:max_size]], axis=1)

        h_max = np.sqrt(np.mean(tmp * tmp) / dim)
        h_est = 10 * h_min
        return h_est, h_min, h_max

    def find_optimal_bandwidth(self, X):
        """Performs maximum likelihood estimation on a test set of the density model fitted on a training set
        """
        from scipy.optimize import fminbound
        X_train, X_test = train_test_split(X, test_size=self.test_ratio_size)
        args = (X_test,)

        hest, hmin, hmax = self.bandwidth_estimate(X_train, X_test)

        print("[kde] Minimum bound = %.4f \t Rough estimate of h = %.4f \t Maximum bound = %.4f" % (hmin, hest, hmax))

        # We are trying to find reasonable tight bounds (hmin, 4.0*hest) to bracket the error function minima
        # Would be nice to have some hard accurate bounds
        self.xtol = round_float(hmin)

        print('[kde] Bandwidth tolerance (xtol) set to precision of minimum bound : %.5f ' % self.xtol)

        self.kde = KernelDensity(algorithm='kd_tree', atol=self.atol, rtol=self.rtol, leaf_size=40, kernel=self.kernel)

        self.kde.fit(X_train)

        # hmax is the upper bound, however, heuristically it appears to always be way above the actual bandwidth. hmax*0.2 seems much better but still conservative
        h_optimal, score_opt, _, niter = fminbound(self.log_likelihood_test_set, hmin, hmax * 0.2, args, maxfun=100,
                                                   xtol=self.xtol, full_output=True)

        print("[kde] Found log-likelihood maximum in %i evaluations, h = %.5f" % (niter, h_optimal))

        if self.extreme_dist is False:  # These bounds should always be satisfied ...
            assert abs(h_optimal - hmax) > 1e-4, "Upper boundary reached for bandwidth"
            assert abs(h_optimal - hmin) > 1e-4, "Lower boundary reached for bandwidth"

        return h_optimal

    # @profile
    def log_likelihood_test_set(self, bandwidth, X_test):
        """Fit the kde model on the training set given some bandwidth and evaluates the negative log-likelihood of the test set
        """
        self.kde.bandwidth = bandwidth
        # l_test = len(X_test)
        return -self.kde.score(X_test[
                               :2000])  # X_test[np.random.choice(np.arange(0, l_test), size=min([int(0.5*l_test), 1000]), replace=False)]) # this should be accurate enough !


def round_float(x):
    """Rounds a float to it's first significant digit"""
    a = list(str(x))
    for i, e in enumerate(a):
        if e != '.':
            if e != '0':
                pos = i
                a[i] = '1'
                break
    return float("".join(a[:pos + 1]))



"""
Methods and functions to compute global descriptors of a frame from its atomic desciptors
"""

import json

import numpy as np

from ..io import NpEncoder


class Atomic_2_Global_Descriptors:
    def __init__(self, k_spec_dict):
        """
        Object handing the reducer functions used to convert atomic descriptors into global ones

        Parameters
        ----------
        k_spec_dict: dictionaries that specify which atomic to global descriptor to use 
        e.g.
        k_spec_dict = {'first_reducer': {'reducer_type': reducer_type,  
                          'zeta': zeta,
                          'species': species,
                          'element_wise': element_wise}}
        """
        self.k_spec_dict = k_spec_dict
        # list of Atomic_2_Global_Descriptor objections
        self.engines = {}

        self.bind()

    def add(self, k_spec, tag):
        """
        adding the specifications of a new Atomic_2_Global_Descriptor
        Parameters
        ----------
        k_spec: a dictionary that specify which Atomic_2_Global_Descriptor descriptor to use 
        """
        self.k_spec_dict[tag] = k_spec

    def pack(self):
        return json.dumps(self.k_spec_dict, sort_keys=True, cls=NpEncoder)

    def bind(self):
        """
        binds the objects that actually compute the descriptors
        these objects need to have .create(atomic_desc) method to compute the global descriptors from atomic ones
        """
        # clear up the objects
        self.engines = {}
        for element in self.k_spec_dict.keys():
            self.engines[element] = self._call(self.k_spec_dict[element])
            self.k_spec_dict[element]['acronym'] = self.engines[element].get_acronym()

    def _call(self, k_spec):
        """
        call the specific descriptor objects
        """
        if "reducer_type" not in k_spec.keys():
            raise ValueError("Did not specify the type of the global descriptor reducer.")
        if k_spec["reducer_type"] == "average":
            return Atomic_2_Global_Average(k_spec)
        if k_spec["reducer_type"] == "sum":
            return Atomic_2_Global_Sum(k_spec)
        if k_spec["reducer_type"] == "moment_average":
            return Atomic_2_Global_Moment_Average(k_spec)
        if k_spec["reducer_type"] == "moment_sum":
            return Atomic_2_Global_Moment_Sum(k_spec)
        else:
            raise NotImplementedError

    def compute(self, atomic_desc_dict, atomic_numbers):
        """
        compute the global descriptor vector for a frame from atomic contributions
        Parameters
        ----------
        atomic_desc_dict : a dictionary. each entry contains the essential info of the descriptor (acronym) 
                          and a np.array [N_desc*N_atoms]. Global descriptors for a frame.
                     see Atomic_Descriptors.compute() in .atomic_descriptors.py
        atomic_numbers: np.matrix. [N_atoms]. Atomic numbers for atoms in the frame.

        Returns
        -------
        desc_dict: a dictionary. each entry contains the essential info of the descriptor, i.e. acronym 
                          and a np.array [N_desc]. Global descriptors for a frame.
                   e.g. {'d1':{ 'acronym': 'XXX', 'descriptors': `a np.array [N_desc]`}}
        """
        desc_dict = {}
        for atomic_desc_element in atomic_desc_dict.keys():
            atomic_desc_now = atomic_desc_dict[atomic_desc_element]['atomic_descriptors']
            desc_dict[atomic_desc_element] = {}
            for element in self.k_spec_dict.keys():
                desc_dict[atomic_desc_element][element] = {}
                k_acronym, desc_dict[atomic_desc_element][element]['descriptors'] = self.engines[element].create(
                    atomic_desc_now, atomic_numbers)
                # we use a combination of the acronym of the descriptor and of the reducer function
                desc_dict[atomic_desc_element][element]['acronym'] = atomic_desc_dict[atomic_desc_element][
                                                                         'acronym'] + k_acronym
        return desc_dict


class Atomic_2_Global_Base:
    def __init__(self, k_spec):
        self.acronym = ""
        # we have defaults here; the default is not to distinguish between different elements
        if 'element_wise' in k_spec.keys():
            self.element_wise = bool(k_spec['element_wise'])
        else:
            self.element_wise = False

        if self.element_wise:
            try:
                self.species = k_spec['species']
            except:
                raise ValueError("Cannot do element-wise operations without specifying the global species")
            self.acronym = "-e"

    def get_acronym(self):
        # we use an acronym for each descriptor, so it's easy to find it and refer to it
        return self.acronym

    def create(self, atomic_desc, atomic_numbers=[]):
        """
        compute the global descriptor vector for a frame from atomic contributions
        Parameters
        ----------
        atomic_desc: a np.array [N_desc*N_atoms]. Atomic descriptors for a frame.
        atomic_numbers: np.matrix. [N_atoms]. Atomic numbers for atoms in the frame.

        Returns
        -------
        acronym: self.acronym
        desc:  a np.array [N_desc]. Global descriptors for a frame.
        """
        return self.acronym, []


class Atomic_2_Global_Average(Atomic_2_Global_Base):
    """this is the vanilla situation. We just take the average soap for all atoms"""

    def __init__(self, k_spec):

        super().__init__(k_spec)

        if "reducer_type" not in k_spec.keys() or k_spec["reducer_type"] != "average":
            raise ValueError("reducer type is not average or cannot find the type")

        print("Using Atomic_2_Global_Average reducer ...")

    def create(self, atomic_desc, atomic_numbers=[]):
        if self.element_wise:
            return self.acronym, Descriptor_By_Species(atomic_desc, atomic_numbers, self.species, True)
        else:
            return self.acronym, np.mean(atomic_desc, axis=0)


class Atomic_2_Global_Sum(Atomic_2_Global_Base):
    """ We just take the sum soap for all atoms"""

    def __init__(self, k_spec):

        super().__init__(k_spec)

        if "reducer_type" not in k_spec.keys() or k_spec["reducer_type"] != "sum":
            raise ValueError("reducer type is not sum or cannot find the type")

        print("Using Atomic_2_Global_Sum reducer ...")
        self.acronym += "-sum"

    def create(self, atomic_desc, atomic_numbers=[]):
        if self.element_wise:
            return self.acronym, Descriptor_By_Species(atomic_desc, atomic_numbers, self.species, False)
        else:
            return self.acronym, np.sum(atomic_desc, axis=0)


class Atomic_2_Global_Moment_Average(Atomic_2_Global_Base):
    """ 
    get the global descriptor from atomic ones 
    by averaging over the atomic descriptors of z th power 

    Parameters
    ----------
    zeta: take the zeta th power
    """

    def __init__(self, k_spec):

        super().__init__(k_spec)

        if "reducer_type" not in k_spec.keys() or k_spec["reducer_type"] != "moment_average":
            raise ValueError("reducer type is not moment_average or cannot find the type")

        try:
            self.zeta = k_spec['zeta']
        except:
            raise ValueError("cannot initialize the zeta value")

        print("Using Atomic_2_Global_Moment_Average reducer ...")
        self.acronym += "-z-" + str(self.zeta)

    def create(self, atomic_desc, atomic_numbers=[]):
        if self.element_wise:
            return self.acronym, Descriptor_By_Species(np.power(atomic_desc, self.zeta), atomic_numbers, self.species,
                                                       True)
        else:
            return self.acronym, np.mean(np.power(atomic_desc, self.zeta), axis=0)


class Atomic_2_Global_Moment_Sum(Atomic_2_Global_Base):
    """ 
    get the global descriptor from atomic ones 
    by averaging over the atomic descriptors of z th power 

    Parameters
    ----------
    zeta: take the zeta th power
    """

    def __init__(self, k_spec):

        super().__init__(k_spec)

        if "reducer_type" not in k_spec.keys() or k_spec["reducer_type"] != "moment_sum":
            raise ValueError("reducer type is not moment_sum or cannot find the type")

        try:
            self.zeta = k_spec['zeta']
        except:
            raise ValueError("cannot initialize the zeta list")

        print("Using Atomic_2_Global_Moment_Sum reducer ...")
        self.acronym += "-z-" + str(self.zeta) + "-sum"

    def create(self, atomic_desc, atomic_numbers=[]):
        if self.element_wise:
            return self.acronym, Descriptor_By_Species(np.power(atomic_desc, self.zeta), atomic_numbers, self.species,
                                                       False)
        else:
            return self.acronym, np.sum(np.power(atomic_desc, self.zeta), axis=0)


def Descriptor_By_Species(atomic_desc, atomic_numbers, global_species, average_over_natom=True):
    """ 
    first compute the average/sum descriptors for each species,
    then concatenate them.

    Parameters
    ----------
    atomic_desc: np.matrix. [N_atoms, N_desc]. Atomic descriptors for a frame.
    atomic_numbers: np.matrix. [N_atoms]. Atomic numbers for atoms in the frame.
    global_species: a list of all atomic species in all frames
    average_over_natom: normalized by number of the atoms of the same species

    Returns
    -------
    desc: np.matrix [N_desc*len(global_species)]. Global descriptors for a frame.
    """
    desc_by_species = {}
    for species in global_species:
        atomic_desc_by_species = [atomic_desc[i] for i, at in enumerate(atomic_numbers) if at == species]
        if average_over_natom and len(atomic_desc_by_species) > 0:
            # normalize by the number of atoms
            desc_by_species[species] = np.mean(atomic_desc_by_species, axis=0)
        elif len(atomic_desc_by_species) > 0:
            desc_by_species[species] = np.sum(atomic_desc_by_species, axis=0)
        else:
            desc_by_species[species] = 0
        # print(np.shape(atomic_desc),len(atomic_desc))

    desc_len = np.shape(atomic_desc)[1]
    desc = np.zeros(desc_len * len(global_species), dtype=float)
    for i, species in enumerate(global_species):
        desc[i * desc_len:(i + 1) * desc_len] = desc_by_species[species]
    return np.asarray(desc)



"""
CUR matrix decomposition is a low-rank matrix decomposition algorithm that is explicitly expressed in a small number of actual columns and/or actual rows of data matrix.
"""

import numpy as np
import scipy.sparse.linalg as spalg
from scipy.linalg import svd, eigh, norm, pinv
from scipy.sparse.linalg import svds


def CUR_deterministic(X, n_col, error_estimate=True, costs=1):
    """
    Given rank k, find the best column and rows
    X = C U R

    Parameters
    ----------
    X: np.matrix
       input a covariance matrix 
    n_col: int
       number of column to keep
    error_estimate: bool, optional
       compute the remaining error of the CUR
    costs: float, optional
        a list of costs associated with each column

    Returns
    -------
    indices : np.array
        indices of columns to choose
    cur_error : np.array
        the error of the decomposition
    """

    RX = X.copy()
    rsel = np.zeros(n_col, dtype=int)
    rerror = np.zeros(n_col, dtype=float)

    for i in range(n_col):
        rval, RX = CUR_deterministic_step(RX, n_col - i, costs)
        rsel[i] = rval
        if error_estimate:
            rerror[i] = np.sum(np.abs(RX))
    return rsel, rerror


def CUR_deterministic_step(cov, k, costs=1):
    """ Apply (deterministic) CUR selection of k rows & columns of the
    given covariance matrix, including an orthogonalization step. Costs can be weighted if desired. """

    evc, evec = spalg.eigs(cov, k)
    # compute the weights and select the one with maximum weight
    weights = np.sum(np.square(evec), axis=1) / costs
    sel = np.argmax(weights)
    vsel = cov[sel] / np.linalg.norm(cov[sel])
    rcov = cov.copy()
    print("selected: ", sel)
    for i in range(len(rcov)):
        # Diagonalize the covariance matrix wrt the chosen column
        rcov[i] -= vsel * np.dot(cov[i].A1, vsel.A1)

    return sel, rcov


def cur_column_select(array, num, rank=None, deterministic=True, mode="sparse", weights=None, calc_error=False):
    """Select columns from a matrix according to statstical leverage score.

    Based on:
    [1] Mahoney MW, Drineas P. CUR matrix decompositions for improved data analysis. PNAS. 2009 Jan 20;106(3):697–702.


    Notes
    -----
    Which mode to use? If the matrix is not square or possibly not hermitian (or even real symmetric) then sparse is
    the only option. For hermitian matrices, the calculation of the eigenvectors with `eigh` may be faster than
    using the sparse svd method. Benchmark is needed for this, especially for descriptor kernels.


    Parameters
    ----------
    array : np.array
        Array to compute find columns of (M, N)
    num : int
        number of column indices to be returned

    rank : int, optional
        number of singular vectors to calculate for calculation of the statistical leverage scores
        default: minimum of (min(N, M) - 1, num/2) but at least one

    deterministic : bool, optional
        Usage of deterministic method or probabilistic.
        Deterministic (True): the top `num` columns are given
        Stochastic (False): use leverage scores as probabilities for choosing num

    mode : {"sparse", "dense", "hermitian"}
        mode of the singular vector calculation.
            - `sparse` (default), which is using sparse-SVD, which is expected to be robust and to solve the problem
            - `hermitian` offers a speedup for hermitian matrices (ie. real symmetric kernel matrices as well)
            - `dense` (not recommended) is using SVD, but calculated all the singular vectors and uses a number of them
            according to the rank parameter

    weights : np.array, optional
        Costs for taking each column. shape=(N,)
        The statistical leverage scores are scaled by this array

    calc_error : bool, optional
        calculate the error of the decomposition, default False
        There is a significant cost to the calculation of this, due to a semi-inverse operation needed.
        The error here is the norm of the matrix of difference between the original and the approximation obtained from
        the selected columns. Not necessarily meaningful in every situation.

    Returns
    -------
    indices : np.array
        indices of columns to choose
    cur_error :
    """

    if not isinstance(rank, int):
        # this is set heuristically
        rank = min(min(array.shape) - 1, max(1, int(num / 2)))

    # compute singular vectors -- the rest is not relevant from this calculation
    if mode == "sparse":
        # left_singular_vectors, _, right_singular_vectors = svds(array, k)  # for error calculation
        *_, right_singular_vectors = svds(array, rank)
    elif mode == "dense":
        *_, right_singular_vectors = svd(array)
        right_singular_vectors = right_singular_vectors[-rank:]
    elif mode == "hermitian":
        assert array.shape[0] == array.shape[1] and len(array.shape) == 2
        _, right_singular_vectors = eigh(array, eigvals=(array.shape[0] - rank, array.shape[0] - 1))
    else:
        raise ValueError(
            'mode {} is not valid for this calculation, use either of: "sparse", "dense", "hermitian"'.format(mode))

    # compute statistical leverage scores and scale by weight if given
    pi_j = np.sum(right_singular_vectors ** 2, axis=0) / rank
    if weights is not None:
        # 1D vector of length equal to the number of columns
        assert np.shape(array)[1] == np.shape(weights)[0] and len(np.shape(weights)) == 1
        pi_j *= weights
        pi_j /= np.sum(pi_j)  # normalise again

    if deterministic:
        indices = np.argsort(pi_j)[-num:]
    else:
        indices = np.random.choice(np.arange(array.shape[1]), size=num, replace=False, p=pi_j)

    if calc_error:
        # error of the representation, ||array -  C dot C+ dot array||
        c = array[:, indices]
        c_pinv = pinv(c)
        cur_error = norm(array - np.matmul(np.matmul(c, c_pinv), array))

        return indices, cur_error
    else:
        return indices



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="asaplib",
    version="0.0.2",
    author="Bingqing Cheng",
    author_email="tonicbq@gmail.com",
    description=
    "Automatic Selection And Prediction tools for materials and molecules",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BingqingCheng/ASAP",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'dscribe>=0.4.0', 'click>=7.0', 'numpy', 'scipy', 'scikit-learn',
        'ase>=3.19.1', 'umap-learn', 'PyYAML', 'tqdm>=4.46.1', 'pandas'
    ],
    extras_require={'testing': ['pytest>=5.0']},
    python_requires='>=3.6',
    # scripts=['scripts/clustering.py',
    #          'scripts/frame_select.py',
    #          'scripts/gen_soap_descriptors.py',
    #          'scripts/kernel_density_estimation.py',
    #          'scripts/kpca.py',
    #          'scripts/kpca_sparse.py',
    #          'scripts/krr.py',
    #          'scripts/pca.py',
    #          'scripts/kpca_for_projection_viewer.py',
    #          'scripts/ridge_regression.py',
    #          'scripts/tsne.py',
    #          'scripts/umap_reducer.py']
    entry_points="""
    [console_scripts]
    asap=asaplib.cli.cmd_asap:asap
    """)
